import unittest

from rpft.parsers.creation.flowparser import FlowParser
from rpft.rapidpro.models.containers import RapidProContainer

from tests.utils import get_table_from_file


class TestFlowParserReverse(unittest.TestCase):
    def setUp(self) -> None:
        pass

    def compare_models_leniently(self, model, model_exp):
        model_dict = model.model_dump()
        model_exp_dict = model_exp.model_dump()
        for edge, edge_exp in zip(model_dict["edges"], model_exp_dict["edges"]):
            if (
                edge["condition"]["variable"] == "@input.text"
                and edge_exp["condition"]["variable"] == ""
            ):
                # @input.text is implicit when left blank.
                edge_exp["condition"]["variable"] = "@input.text"
        if model.node_uuid and not model_exp.node_uuid:
            # If no explicit node UUID was specified, don't compare the generated node
            # uuid
            model_dict["node_uuid"] = ""
        if model.obj_id and not model_exp.obj_id:
            # If no explicit object UUID was specified, don't compare
            model_dict["obj_id"] = ""
        for field in ["obj_name", "node_name", "ui_type"]:
            # We ignore these. Attachments need to be rethought,
            # the other of these fields may be removed as well.
            model_dict.pop(field)
            model_exp_dict.pop(field)
        # TODO: We have trailing '' quick replies and webhook.headers here.
        # They get removed in the conversion process, but probably the RowParser
        # should already take care of them
        model_exp_dict["choices"] = [qr for qr in model_exp_dict["choices"] if qr]
        model_exp_dict["webhook"]["headers"] = [
            header for header in model_exp_dict["webhook"]["headers"] if header
        ]
        self.assertEqual(model_dict, model_exp_dict)

    def run_example(self, filename, flow_name):
        self.maxDiff = None
        table = get_table_from_file(filename)
        container = RapidProContainer()
        parser = FlowParser(container, flow_name, table)
        flow = parser.parse()
        container.validate()
        row_models = flow.to_rows(numbered=True)
        data_rows = parser.sheet_parser.parse_all()
        self.assertEqual(len(row_models), len(data_rows))
        for model, model_exp in zip(row_models, data_rows):
            self.compare_models_leniently(model, model_exp)

    def test_no_switch_nodes(self):
        self.run_example("input/no_switch_nodes.csv", "no_switch_nodes")

    def test_switch_nodes(self):
        self.run_example("input/switch_nodes.csv", "switch_nodes")

    def test_loop_from_start(self):
        self.run_example("input/loop_from_start.csv", "loop_from_start")

    def test_groups_and_flows(self):
        self.run_example("input/groups_and_flows.csv", "groups_and_flows")

    def test_loop_and_multiple_conditions(self):
        self.run_example(
            "input/loop_and_multiple_conditions.csv", "loop_and_multiple_conditions"
        )

    def test_rejoin(self):
        self.run_example("input/rejoin.csv", "rejoin")



import importlib
import logging
from collections import OrderedDict

from rpft.logger.logger import logging_context
from rpft.parsers.creation import globalrowmodels
from rpft.parsers.creation.campaigneventrowmodel import CampaignEventRowModel
from rpft.parsers.creation.campaignparser import CampaignParser
from rpft.parsers.creation.contentindexrowmodel import (
    ContentIndexRowModel,
    ContentIndexType,
)
from rpft.parsers.creation.flowparser import FlowParser
from rpft.parsers.creation.models import ChatbotDefinition, TemplateSheet
from rpft.parsers.creation.tagmatcher import TagMatcher
from rpft.parsers.creation.surveyparser import Survey, SurveyParser, SurveyQuestion
from rpft.parsers.creation.triggerparser import TriggerParser
from rpft.parsers.creation.triggerrowmodel import TriggerRowModel
from rpft.rapidpro.models.containers import RapidProContainer


LOGGER = logging.getLogger(__name__)


class DataSheet:
    def __init__(self, rows, row_model):
        """Args:
        rows: A dict mapping row_ids (str) to row_model instances.
        row_model: the model underlying the instances of rows.
        """
        self.rows = rows
        self.row_model = row_model

    def to_dict(self):
        return {
            "model": self.row_model.__name__,
            "rows": [content.model_dump() for content in self.rows.values()],
        }


class ParserError(Exception):
    pass


class ContentIndexParser:

    def __init__(
        self,
        data_source,
        user_data_model_module_name=None,
        tag_matcher=TagMatcher(),
    ):
        self.data_source = data_source
        self.tag_matcher = tag_matcher
        self.template_sheets = {}
        self.data_sheets = {}
        self.flow_definition_rows = []
        self.campaign_parsers: dict[str, tuple[str, CampaignParser]] = {}
        self.surveys = {}
        self.survey_questions: list[SurveyQuestion] = []
        self.trigger_parsers = OrderedDict()
        self.global_context = {}
        self.user_models_module = (
            importlib.import_module(user_data_model_module_name)
            if user_data_model_module_name
            else None
        )
        indices = self.data_source.get_all("content_index", ContentIndexRowModel)

        if not indices:
            raise Exception("No content index found")

        for entries, location, key in indices:
            self._process_content_index_table(entries, f"{location}-{key}")

        self._populate_missing_templates()
        self.definition = ChatbotDefinition(
            self.flow_definition_rows,
            self.data_sheets,
            self.template_sheets,
            self.surveys,
            self.survey_questions,
            {"globals": self.global_context},
        )

    def _process_content_index_table(self, rows, label):
        for row_idx, row in enumerate(rows, start=2):
            logging_prefix = f"{label} | row {row_idx}"

            with logging_context(logging_prefix):
                if row.status == "draft":
                    continue

                if not self.tag_matcher.matches(row.tags):
                    continue

                if len(row.sheet_name) != 1 and row.type not in [
                    ContentIndexType.DATA_SHEET.value,
                    ContentIndexType.SURVEY.value,
                    ContentIndexType.SURVEY_QUESTION.value,
                ]:
                    raise Exception(
                        f"For {row.type} rows, exactly one sheet_name has to be"
                        " specified"
                    )

                if row.type == ContentIndexType.CONTENT_INDEX.value:
                    entries, location, key = self.data_source.get(
                        row.sheet_name[0], ContentIndexRowModel
                    )

                    with logging_context(f"{key}"):
                        self._process_content_index_table(entries, f"{location}-{key}")
                elif row.type == ContentIndexType.DATA_SHEET.value:
                    if not len(row.sheet_name) >= 1:
                        raise Exception(
                            "For data_sheet rows, at least one sheet_name has to be"
                            " specified"
                        )

                    self._process_data_sheet(row)
                elif row.type == ContentIndexType.TEMPLATE.value:
                    if row.new_name:
                        LOGGER.warning(
                            "template_definition does not support 'new_name'; "
                            f"new_name '{row.new_name}' will be ignored."
                        )

                    self._add_template(row, True)
                elif row.type == ContentIndexType.FLOW.value:
                    self.flow_definition_rows.append((logging_prefix, row))
                elif row.type == ContentIndexType.CAMPAIGN.value:
                    campaign_parser = self.create_campaign_parser(row)
                    name = campaign_parser.campaign.name

                    if name in self.campaign_parsers:
                        LOGGER.debug(
                            f"Duplicate campaign definition sheet '{name}'. "
                            "Overwriting previous definition."
                        )

                    self.campaign_parsers[name] = (logging_prefix, campaign_parser)
                elif row.type == ContentIndexType.TRIGGERS.value:
                    self.trigger_parsers[row.sheet_name[0]] = (
                        logging_prefix,
                        self.create_trigger_parser(row),
                    )
                elif row.type == ContentIndexType.SURVEY.value:
                    self._add_survey(row, logging_prefix)
                elif row.type == ContentIndexType.SURVEY_QUESTION.value:
                    self._add_survey_question(row, logging_prefix)
                elif row.type == ContentIndexType.GLOBALS.value:
                    self._process_globals_sheet(row)
                elif row.type == ContentIndexType.IGNORE.value:
                    self._process_ignore_row(row.sheet_name[0])
                else:
                    LOGGER.error(f"invalid type: '{row.type}'")

    def _add_template(self, row, update_duplicates=False):
        sheet_name = row.sheet_name[0]

        if sheet_name in self.template_sheets and update_duplicates:
            LOGGER.debug(
                f"Duplicate template definition sheet '{sheet_name}'. "
                "Overwriting previous definition."
            )

        if sheet_name not in self.template_sheets or update_duplicates:
            sheet = self.data_source._get_sheet_or_die(sheet_name)
            self.template_sheets[sheet_name] = TemplateSheet(
                sheet_name,
                sheet.table,
                row.template_argument_definitions,
            )

    def _process_ignore_row(self, sheet_name):
        self.flow_definition_rows = [
            (logging_prefix, row)
            for logging_prefix, row in self.flow_definition_rows
            if (row.new_name or row.sheet_name[0]) != sheet_name
        ]
        self.campaign_parsers.pop(sheet_name, None)
        self.trigger_parsers.pop(sheet_name, None)
        self.surveys.pop(sheet_name, None)

    def _populate_missing_templates(self):
        for logging_prefix, row in self.flow_definition_rows:
            with logging_context(f"{logging_prefix} | {row.sheet_name[0]}"):
                self._add_template(row)

    def _process_globals_sheet(self, row):
        properties, *_ = self.data_source.get(
            row.sheet_name[0],
            globalrowmodels.IDValueRowModel,
        )
        context_dict = {r.ID: r.value for r in properties}
        intersection = self.global_context.keys() & context_dict.keys()
        if intersection:
            LOGGER.info(f"Overwriting globals {intersection}")
        self.global_context |= context_dict

    def _process_data_sheet(self, row):
        sheet_names = row.sheet_name

        if row.operation.type in ["filter", "sort"] and len(sheet_names) > 1:
            LOGGER.warning(
                "data_sheet definition take only one sheet_name for filter and sort "
                "operations. All but the first sheet_name are ignored."
            )

        if not row.operation.type:
            if len(sheet_names) > 1:
                LOGGER.warning(
                    "Implicitly concatenating data sheets without concat operation "
                    "is deprecated and may be removed in the future."
                )
                data_sheet = self._data_sheets_concat(sheet_names, row.data_model)
            else:
                data_sheet = self._get_new_data_sheet(sheet_names[0], row.data_model)

        else:
            if not row.new_name:
                raise Exception(
                    "If an operation is applied to a data_sheet, a new_name has to be"
                    " provided"
                )

            if row.operation.type == "concat":
                data_sheet = self._data_sheets_concat(sheet_names, row.data_model)
            elif row.operation.type == "filter":
                data_sheet = self._data_sheets_filter(
                    sheet_names[0], row.data_model, row.operation
                )
            elif row.operation.type == "sort":
                data_sheet = self._data_sheets_sort(
                    sheet_names[0], row.data_model, row.operation
                )
            else:
                raise Exception(f'Unknown operation "{row.operation}"')

        new_name = row.new_name or sheet_names[0]

        if new_name in self.data_sheets:
            LOGGER.debug(
                f"Duplicate data sheet {new_name}. Overwriting previous definition."
            )

        self.data_sheets[new_name] = data_sheet

    def _get_data_sheet(self, sheet_name, data_model_name):
        if sheet_name in self.data_sheets:
            return self.data_sheets[sheet_name]
        else:
            return self._get_new_data_sheet(sheet_name, data_model_name)

    def _get_new_data_sheet(self, sheet_name, data_model_name=None):
        model = (
            getattr(self.user_models_module, data_model_name, None)
            or getattr(globalrowmodels, data_model_name, None)
            if data_model_name
            else None
        )

        if data_model_name and not model:
            raise Exception(
                f'Undefined data_model_name "{data_model_name}" '
                f"in {self.user_models_module}."
            )

        with logging_context(sheet_name):
            items, *_ = self.data_source.get(sheet_name, model)
            model_instances = OrderedDict((item.ID, item) for item in items)

            return DataSheet(model_instances, model)

    def _data_sheets_concat(self, sheet_names, data_model_name):
        all_data_rows = OrderedDict()
        user_model = None

        for sheet_name in sheet_names:
            with logging_context(sheet_name):
                data_sheet = self._get_data_sheet(sheet_name, data_model_name)

                if user_model and user_model is not data_sheet.row_model:
                    raise Exception(
                        "Cannot concatenate data_sheets with different underlying"
                        " models"
                    )

                user_model = data_sheet.row_model
                all_data_rows.update(data_sheet.rows)

        return DataSheet(all_data_rows, user_model)

    def _data_sheets_filter(self, sheet_name, data_model_name, operation):
        data_sheet = self._get_data_sheet(sheet_name, data_model_name)
        new_row_data = OrderedDict()

        for row_id, row in data_sheet.rows.items():
            try:
                if eval(operation.expression, {}, dict(row)) is True:
                    new_row_data[row_id] = row
            except NameError as e:
                raise Exception(f"Invalid filtering expression: {e}")
            except SyntaxError as e:
                raise Exception(
                    f'Invalid filtering expression: "{e.text}". '
                    f"SyntaxError at line {e.lineno} character {e.offset}"
                )

        return DataSheet(new_row_data, data_sheet.row_model)

    def _data_sheets_sort(self, sheet_name, data_model_name, operation):
        data_sheet = self._get_data_sheet(sheet_name, data_model_name)

        try:
            new_row_data = OrderedDict(
                sorted(
                    data_sheet.rows.items(),
                    key=lambda kvpair: eval(operation.expression, {}, dict(kvpair[1])),
                    reverse=operation.order.lower() == "descending",
                )
            )
        except NameError as e:
            raise Exception(f"Invalid sorting expression: {e}")
        except SyntaxError as e:
            raise Exception(
                f'Invalid sorting expression: "{e.text}". '
                f"SyntaxError at line {e.lineno} character {e.offset}"
            )

        return DataSheet(new_row_data, data_sheet.row_model)

    def data_sheets_to_dict(self):
        sheets = {}

        for sheet_name, sheet in self.data_sheets.items():
            sheets[sheet_name] = sheet.to_dict()

        return {
            "sheets": sheets,
            "meta": {
                "user_models_module": self.user_models_module.__name__,
                "version": "0.1.0",
            },
        }

    def parse_all(self):
        rapidpro_container = RapidProContainer()
        self.parse_all_flows(rapidpro_container)
        self.parse_all_campaigns(rapidpro_container)
        self.parse_all_triggers(rapidpro_container)
        self.parse_all_surveys(rapidpro_container)

        return rapidpro_container

    def create_campaign_parser(self, row):
        sheet_name = row.sheet_name[0]
        rows, *_ = self.data_source.get(sheet_name, CampaignEventRowModel)

        return CampaignParser(row.new_name or sheet_name, row.group, rows)

    def create_trigger_parser(self, row):
        sheet_name = row.sheet_name[0]
        rows, *_ = self.data_source.get(sheet_name, TriggerRowModel)

        return TriggerParser(sheet_name, rows)

    def _add_survey(self, row, logging_prefix):
        name = row.new_name or row.data_sheet

        with logging_context(logging_prefix):
            if name in self.surveys:
                LOGGER.warning(
                    f"Duplicate survey definition sheet '{name}'. "
                    "Overwriting previous definition."
                )

        self.surveys[name] = Survey(
            name,
            self.data_sheets[row.data_sheet],
            row.survey_config,
            row.template_arguments,
            logging_prefix,
        )

    def _add_survey_question(self, row, logging_prefix):
        survey_name = row.new_name or row.data_sheet
        data_row = self.data_sheets[row.data_sheet].rows.get(row.data_row_id)
        if not data_row:
            with logging_context(logging_prefix):
                LOGGER.error(
                    f"No data_row_id given, or data_row_id '{row.data_row_id}' not"
                    f" present in survey question definition sheet '{row.data_sheet}'."
                    " Omitting survey question."
                )
            return

        self.survey_questions.append(
            SurveyQuestion(
                survey_name,
                data_row,
                row.template_arguments,
                logging_prefix,
            )
        )

    def parse_all_campaigns(self, rapidpro_container):
        for logging_prefix, campaign_parser in self.campaign_parsers.values():
            sheet_name = campaign_parser.campaign.name

            with logging_context(f"{logging_prefix} | {sheet_name}"):
                campaign = campaign_parser.parse()
                rapidpro_container.add_campaign(campaign)

    def parse_all_surveys(self, rapidpro_container):
        SurveyParser.parse_all(self.definition, rapidpro_container)

    def parse_all_triggers(self, rapidpro_container):
        for logging_prefix, trigger_parser in self.trigger_parsers.values():
            sheet_name = trigger_parser.sheet_name

            with logging_context(f"{logging_prefix} | {sheet_name}"):
                triggers = trigger_parser.parse()

                for trigger in triggers:
                    rapidpro_container.add_trigger(trigger)

    def parse_all_flows(self, rapidpro_container):
        FlowParser.parse_all(self.definition, rapidpro_container)



import copy

from rpft.parsers.common.rowdatasheet import RowDataSheet
from rpft.parsers.common.rowparser import RowParser
from rpft.logger.logger import logging_context


class SheetParser:
    def parse_sheet(table, row_model):
        """
        Args:
            table: Tablib Dataset representing the table to be parsed.
            row_model: Data model to convert rows of the sheet into.

        Returns:
            RowDataSheet instance containing a list of row_model instances
        """

        sheet_parser = SheetParser(table, row_model)
        return sheet_parser.get_row_data_sheet()

    def __init__(self, table, row_model=None, row_parser=None, context={}):
        """
        Either a row_parser or a row_model need to be provided.

        Args:
            table: Tablib Dataset representing the table to be parsed.
            row_model: Data model to convert rows of the sheet into.
            row_parser: parser to convert flat dicts to RowModel instances.
            context: context used for template parsing
        """

        if not (row_parser or row_model):
            raise ValueError("SheetParser: needs either row_parser or row_model")
        self.row_parser = row_parser or RowParser(row_model)
        self.bookmarks = {}
        self.input_rows = []
        for row_idx, row in enumerate(table):
            row_dict = {h: e for h, e in zip(table.headers, row)}
            self.input_rows.append((row_dict, row_idx + 2))
        self.iterator = iter(self.input_rows)
        self.context = copy.deepcopy(context)

    def add_to_context(self, key, value):
        self.context[key] = value

    def remove_from_context(self, key):
        self.context.pop(key, None)

    def create_bookmark(self, name):
        self.bookmarks[name] = copy.copy(self.iterator)

    def go_to_bookmark(self, name):
        self.iterator = copy.copy(self.bookmarks[name])

    def remove_bookmark(self, name):
        self.bookmarks.pop(name)

    def parse_next_row(self, omit_templating=False, return_index=False):
        try:
            input_row, row_idx = next(self.iterator)
        except StopIteration:
            return (None, None) if return_index else None
        context = self.context if not omit_templating else None
        with logging_context(f"row {row_idx}"):
            row = self.row_parser.parse_row(input_row, context)
        return (row, row_idx) if return_index else row

    def parse_all(self):
        self.iterator = iter(self.input_rows)
        output_rows = []
        row = self.parse_next_row()
        while row is not None:
            output_rows.append(row)
            row = self.parse_next_row()
        return output_rows

    def get_row_data_sheet(self):
        rows = self.parse_all()
        return RowDataSheet(self.row_parser, rows)



from rpft.parsers.creation.flowrowmodel import FlowRowModel


def get_start_row():
    return FlowRowModel(
        row_id="1",
        edges=[{"from_": "start"}],
        type="send_message",
        mainarg_message_text="Text of message",
        choices=["Answer 1", "Answer 2"],
    )


def get_message_with_templating():
    return FlowRowModel(
        row_id="1",
        edges=[{"from_": "start"}],
        type="send_message",
        mainarg_message_text="Default text of message",
        wa_template={
            "name": "template name",
            "uuid": "template uuid",
            "variables": ["var1", "var2"],
        },
    )


def get_unconditional_node_from_1():
    return FlowRowModel(
        row_id="2",
        type="send_message",
        edges=[{"from_": "1"}],
        mainarg_message_text="Unconditional message",
    )


def get_conditional_node_from_1():
    return FlowRowModel(
        row_id="3",
        type="send_message",
        edges=[
            {
                "from_": "1",
                "condition": {
                    "value": "3",
                    "variable": "@fields.name",
                    "type": "has_phrase",
                    "name": "3",
                },
            }
        ],
        mainarg_message_text="Message if @fields.name == 3",
    )



import json
import os

from google_auth_oauthlib.flow import InstalledAppFlow
from googleapiclient.discovery import build
from google.auth.transport.requests import Request
from google.oauth2.credentials import Credentials
from google.oauth2.service_account import Credentials as ServiceAccountCredentials
from datetime import datetime


EXT_MIME_TYPE = {
    ".ods": "application/vnd.oasis.opendocument.spreadsheet",
    ".xlsx": "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
}

# If modifying these scopes, delete the file token.json.
SCOPES = [
    "https://www.googleapis.com/auth/drive.readonly",
    "https://www.googleapis.com/auth/spreadsheets.readonly",
]

ALL_SCOPES = SCOPES + [
    "https://www.googleapis.com/auth/devstorage.read_write",
]


def get_credentials(scopes = None):
    scopes = scopes or SCOPES

    sa_creds = os.getenv("CREDENTIALS")

    if sa_creds:
        return ServiceAccountCredentials.from_service_account_info(
            json.loads(sa_creds), scopes=scopes
        )

    creds = None
    token_file_name = "token.json"

    if os.path.exists(token_file_name):
        creds = Credentials.from_authorized_user_file(
            token_file_name,
            scopes=scopes,
        )

    # If there are no (valid) credentials available, let the user log in.
    if not creds or not creds.valid:
        if creds and creds.expired and creds.refresh_token:
            creds.refresh(Request())
        else:
            flow = InstalledAppFlow.from_client_secrets_file("credentials.json", ALL_SCOPES)
            creds = flow.run_local_server(port=0)

        # Save the credentials for the next run
        with open(token_file_name, "w") as token:
            token.write(creds.to_json())

        # Load the credentials with only the scopes needed for this task
        creds = Credentials.from_authorized_user_file(
            token_file_name,
            scopes=scopes,
        )

    return creds


class Drive:

    @classmethod
    def client(cls):
        if not getattr(cls, "_client", None):
            cls._client = build("drive", "v3", credentials=get_credentials())

        return cls._client

    @classmethod
    def meta(cls, file_id):
        return (
            cls.client()
            .files()
            .get(
                fileId=file_id,
                supportsAllDrives=True,
            )
            .execute()
        )

    @classmethod
    def fetch(cls, file_id):
        meta = cls.meta(file_id)
        content = (
            cls.client()
            .files()
            .get_media(
                fileId=file_id,
                supportsAllDrives=True,
            )
            .execute()
        )

        return meta["name"], content

    @classmethod
    def export(cls, file_id, ext=None, mime_type=None):
        try:
            mime = mime_type if mime_type else EXT_MIME_TYPE[ext]
        except KeyError:
            raise ValueError("Failed to determine MIME type for export")

        meta = cls.meta(file_id)
        content = (
            cls.client()
            .files()
            .export(
                fileId=file_id,
                mimeType=mime,
            )
            .execute()
        )

        return meta["name"], content

    @classmethod
    def get_modified_time(cls, file_id_ls):
        if type(file_id_ls) is str:
            file_id_ls = [file_id_ls]

        modified_time_dict = {}

        def modified_time_callback(request_id, response, exception):
            if exception is None:
                file_id = response.get('id')
                modified_time_str = response.get('modifiedTime')
                if modified_time_str.endswith("Z"):
                    modified_time_str = modified_time_str[:-1] + "+00:00"
                modified_time_dict[file_id] = datetime.fromisoformat(modified_time_str)
            else:
                raise exception
            
        batch = cls.client().new_batch_http_request(callback=modified_time_callback)

        for file_id in file_id_ls:
            batch.add(cls.client().files().get(fileId=file_id, 
                fields='id,modifiedTime', supportsAllDrives=True)
            )

        batch.execute()

        return modified_time_dict
        


import copy
from itertools import zip_longest

from rpft.rapidpro.models.common import FlowReference, Group


class RapidProTriggerError(Exception):
    pass


class Trigger:
    def __init__(
        self,
        trigger_type,
        keywords=None,
        channel=None,
        match_type=None,
        flow=None,
        flow_name=None,
        flow_uuid=None,
        groups=None,
        group_names=None,
        group_uuids=None,
        exclude_groups=None,
        exclude_group_names=None,
        exclude_group_uuids=None,
    ):
        self.trigger_type = trigger_type
        self.keywords = keywords or []
        self.match_type = match_type or None
        if self.trigger_type == "K":
            if not keywords or not keywords[0]:
                raise ValueError('Triggers of type "K" must have a keyword')
            if not match_type:
                self.match_type = "F"
        self.channel = channel or None
        if not flow and not flow_name:
            raise ValueError("Trigger must have flow or a flow_name")
        self.flow = flow or FlowReference(flow_name, flow_uuid)
        if groups is not None:
            self.groups = groups
        else:
            self.groups = []
            self._assign_groups(self.groups, group_names, group_uuids)
        if exclude_groups is not None:
            self.exclude_groups = exclude_groups
        else:
            self.exclude_groups = []
            if exclude_group_names is not None:
                self._assign_groups(
                    self.exclude_groups,
                    exclude_group_names,
                    exclude_group_uuids,
                )

    def _assign_groups(self, groups_field, group_names, group_uuids):
        for group_name, group_uuid in zip_longest(group_names, group_uuids):
            if not group_name:
                raise ValueError("Trigger group must have a name.")
            group = Group(group_name, group_uuid or None)
            groups_field.append(group)

    def from_dict(data):
        data_copy = copy.deepcopy(data)
        if "flow" in data_copy:
            data_copy["flow"] = FlowReference(**data_copy["flow"])
        groups = []
        for group in data_copy["groups"]:
            groups.append(Group.from_dict(group))
        data_copy["groups"] = groups
        if "exclude_groups" in data_copy:
            exclude_groups = []
            for group in data_copy["exclude_groups"]:
                exclude_groups.append(Group.from_dict(group))
            data_copy["exclude_groups"] = exclude_groups
        # Newer versions of RapidPro use a list of keywords instead of a single keyword
        # If we have "keywords", use that. Otherwise, process the "keyword" entry
        if "keywords" not in data_copy:
            assert "keyword" in data_copy
            data_copy["keywords"] = []
            if data_copy["keyword"] is not None:
                data_copy["keywords"].append(data_copy["keyword"])
        if "keyword" in data_copy:
            data_copy.pop("keyword")
        return Trigger(**data_copy)

    def record_global_uuids(self, uuid_dict, require_existing=False):
        if require_existing:
            if not uuid_dict.contains_flow(self.flow.name):
                raise RapidProTriggerError(
                    f"Trigger references undefined flow name {self.flow.name}"
                )
        if self.flow is not None:
            self.flow.record_uuid(uuid_dict)
        if self.groups is not None:
            for group in self.groups:
                group.record_uuid(uuid_dict)
        if self.exclude_groups is not None:
            for group in self.exclude_groups:
                group.record_uuid(uuid_dict)

    def assign_global_uuids(self, uuid_dict):
        if self.flow is not None:
            self.flow.assign_uuid(uuid_dict)
        if self.groups is not None:
            for group in self.groups:
                group.assign_uuid(uuid_dict)
        if self.exclude_groups is not None:
            for group in self.exclude_groups:
                group.assign_uuid(uuid_dict)

    def render(self):
        # We include both keywords and keyword in the render output
        # in order to support both newer and older versions of RapidPro
        render_dict = {
            "trigger_type": self.trigger_type,
            "keyword": self.keywords[0] if self.keywords else None,
            "keywords": self.keywords,
            "channel": self.channel,
            "flow": self.flow.render(),
            "groups": [group.render() for group in self.groups],
            "exclude_groups": [group.render() for group in self.exclude_groups],
        }
        if self.match_type:
            render_dict.update({"match_type": self.match_type})
        return render_dict



import re


class Context(object):
    def __init__(
        self, group_names=None, inputs=None, random_choices=None, variables=None
    ):
        if group_names is not None:
            self.group_names = group_names
        else:
            self.group_names = []

        if inputs is not None:
            self.inputs = inputs
        else:
            self.inputs = []

        if random_choices is not None:
            self.random_choices = random_choices
        else:
            self.random_choices = []

        if variables is not None:
            self.variables = variables
        else:
            self.variables = dict()


def find_node_by_uuid(flow, node_uuid):
    """Given a node uuid, finds the corresponding node.

    Args:
        flow: flow to search in
        node_uuid: node to search for

    Returns:
        node with uuid matchign node_uuid
        None if no node was found.
    """

    for node in flow["nodes"]:
        if node["uuid"] == node_uuid:
            return node
    return None


def find_destination_uuid(current_node, context):
    """
    For a given node, find the next node that is visited and return its uuid.

    The groups the user is in may affect the outcome.

    Args:
        current_node:
        group_names (`list` of `str`): groups the user is in

    Returns:
        uuid of the node visited after this node.
        Maybe be None if it is the last node.
    """

    # By default, choose the first exit.
    destination_uuid = current_node["exits"][0].get("destination_uuid", None)
    if "router" in current_node:
        router = current_node["router"]
        if router["type"] == "switch":
            # Get value of the operand
            if router["operand"] == "@contact.groups":
                operand = context.group_names
            elif router["operand"] == "@input.text" or current_node["actions"]:
                # Actions implies CallWebhook, TransferAirtime or EnterFlow
                operand = context.inputs.pop(0)
            elif router["operand"] in context.variables:
                operand = context.variables[router["operand"]]
            else:
                operand = None

            # Find the case that applies here and get its category_uuid
            # The arguments are not parse (e.g. no variable substitution)
            category_uuid = router[
                "default_category_uuid"
            ]  # The "Other" option (default)
            if operand is None:
                # Input "None" indicates No Response if the router has a wait.
                if "wait" in router:
                    if (
                        "timeout" in router["wait"]
                        and router["wait"]["timeout"]["seconds"] > 0
                    ):
                        category_uuid = router["wait"]["timeout"]["category_uuid"]
                    else:
                        return None  # We're stuck here forever --> exit here.
            else:
                for case in router["cases"]:
                    if case["type"] == "has_group":
                        group_name = case["arguments"][1]
                        if group_name in operand:
                            # Note: We ignore the Group UUID here.
                            category_uuid = case["category_uuid"]
                            break
                    elif case["type"] == "has_phrase":
                        if case["arguments"][0].lower() in operand.lower():
                            category_uuid = case["category_uuid"]
                            break
                    elif (
                        case["type"] == "has_only_text"
                        or case["type"] == "has_category"
                    ):
                        # These are case sensitive
                        if case["arguments"][0] == operand:
                            category_uuid = case["category_uuid"]
                            break
                    elif case["type"] == "has_any_word":
                        matched = False
                        for word in case["arguments"][0].split():
                            for input_word in operand.split():
                                if word.lower() == input_word.lower():
                                    matched = True
                        if matched:
                            category_uuid = case["category_uuid"]
                            break
                    elif case["type"] == "has_text":
                        if case["arguments"] != []:
                            raise ValueError(
                                "has_text case type must not have arguments"
                            )
                        if operand.strip() != "":
                            category_uuid = case["category_uuid"]
                            break
                    elif case["type"] == "has_email":
                        if case["arguments"] != []:
                            raise ValueError(
                                "has_email case type must not have arguments"
                            )
                        if len(re.findall(r"[\w]+@[\w]+\.[\w]+", operand)) > 0:
                            # This might differ from the RapidPro implementation.
                            category_uuid = case["category_uuid"]
                            break
                    elif case["type"] == "has_number_between":
                        # This might differ from the RapidPro implementation.
                        if len(case["arguments"]) != 2:
                            raise ValueError("has_number_between must have 2 arguments")
                        try:
                            number = float(operand)
                        except ValueError:
                            number = None
                        if number is not None and float(
                            case["arguments"][0]
                        ) <= number <= float(case["arguments"][1]):
                            category_uuid = case["category_uuid"]
                            break
                    elif case["type"] == "has_number_lt":
                        # This might differ from the RapidPro implementation.
                        try:
                            number = float(operand)
                        except ValueError:
                            number = None
                        if number is not None and number < float(case["arguments"][0]):
                            category_uuid = case["category_uuid"]
                            break
                    elif case["type"] == "has_number_gt":
                        # This might differ from the RapidPro implementation.
                        try:
                            number = float(operand)
                        except ValueError:
                            number = None
                        if number is not None and number > float(case["arguments"][0]):
                            category_uuid = case["category_uuid"]
                            break

            # Find the category matching the case and get its exit_uuid
            exit_uuid = None
            for category in router["categories"]:
                if category["uuid"] == category_uuid:
                    exit_uuid = category["exit_uuid"]
            if exit_uuid is None:
                raise ValueError(
                    "No valid exit_uuid in router of node with uuid "
                    + current_node["uuid"]
                )
        else:  # router["type"] == "random"
            # Get the exit_uuid from a random category
            random_choice = context.random_choices.pop(0)
            exit_uuid = router["categories"][random_choice]["exit_uuid"]

        # Find the exit we take here and get its destination_uuid
        destination_uuid = (
            -1
        )  # -1 because None is a valid value indicating the end of a flow
        for exit in current_node["exits"]:
            if exit["uuid"] == exit_uuid:
                destination_uuid = exit.get("destination_uuid", None)
                break
        if destination_uuid == -1:
            raise ValueError(
                "No valid destination_uuid in router of node with uuid "
                + current_node["uuid"]
            )
    return destination_uuid


# List of actions: https://app.rapidpro.io/mr/docs/flows.html#actions
action_value_fields = {
    "add_contact_groups": (lambda x: x["groups"][0]["name"]),
    "add_contact_urn": (lambda x: x["path"]),
    "add_input_labels": (lambda x: x["labels"][0]["name"]),
    "call_classifier": (lambda x: x["classified"]["name"]),
    "call_resthook": (lambda x: x["resthook"]),
    "call_webhook": (lambda x: x["url"]),
    "enter_flow": (lambda x: x["flow"]["name"]),
    "open_ticket": (lambda x: x["subject"]),
    "play_audio": (lambda x: x["audio_url"]),
    "remove_contact_groups": (
        lambda x: x["groups"][0]["name"] if x["groups"] else "ALL"
    ),
    "say_msg": (lambda x: x["text"]),
    "send_broadcast": (lambda x: x["text"]),
    "send_email": (lambda x: x["subject"]),
    "send_msg": (lambda x: x["text"]),
    "set_contact_channel": (lambda x: x["channel"]["name"]),
    "set_contact_field": (lambda x: x["field"]["key"]),
    "set_contact_language": (lambda x: x["language"]),
    "set_contact_name": (lambda x: x["name"]),
    "set_contact_status": (lambda x: x["status"]),
    "set_contact_timezone": (lambda x: x["timezone"]),
    "set_run_result": (lambda x: x["name"]),
    "start_session": (lambda x: x["flow"]["name"]),
    "transfer_airtime": (lambda x: "Amount"),
}


def process_actions(node, context):
    """May modify the context."""

    outputs = []
    for action in node["actions"]:
        # Log the action, regardless of its type
        # TODO: Try/catch in case of unrecognized action/missing field?
        action_type = action["type"]
        action_value = action_value_fields[action_type](action)
        outputs.append((action_type, action_value))

        # We only support a very small subset of actions.
        if action_type == "add_contact_groups":
            for group in action["groups"]:
                context.group_names.append(group["name"])
        elif action_type == "enter_flow":
            # TODO: recurse, append outputs
            # We need to have all flows for that.
            # action["flow"]["name"]
            # action["flow"]["uuid"]
            pass
    return outputs


def traverse_flow(flow, context):
    """
    Traverse a given flow, assuming the user's group memberships
    as specified in group_names, which determine which path through
    the flow is taken.

    Traversal ends once we reach an exit with destination_uuid None.

    If we encounter a destination_uuid leading to a node not contained
    in the flow, the flow is considered erroneous and an error is raised.

    Returns:
        A list of strings, which are the outputs of send_msg actions
        that are encounters while traversing through the flow.

    Only supports send_msg actions and group switches
    TODO: Abort after too many steps (there may be cycles).
    """

    outputs = []
    current_node = flow["nodes"][0]
    while current_node is not None:
        outputs += process_actions(current_node, context)

        destination_uuid = find_destination_uuid(current_node, context)
        if destination_uuid is None:  # we've reached the exit
            break
        current_node = find_node_by_uuid(flow, destination_uuid)
        if current_node is None:
            raise ValueError("Destination_uuid {} is invalid.".format(destination_uuid))
    return outputs


def find_final_destination(flow, node, context):
    """Starting at node in flow, traverse the flow until we reach a
    destination that is not contained inside the flow.
    TODO: Abort after too many steps (there may be cycles).

    Returns:
        uuid of the destination outside the flow
    """

    while node is not None:
        process_actions(node, context)
        destination_uuid = find_destination_uuid(node, context)
        node = find_node_by_uuid(flow, destination_uuid)
    return destination_uuid



from pydantic import field_validator, model_validator

from rpft.parsers.common.rowparser import ParserModel


class TriggerRowModel(ParserModel):
    type: str
    keywords: list[str] = []
    flow: str = ""
    groups: list[str] = []
    exclude_groups: list[str] = []
    channel: str = ""
    match_type: str = ""

    @field_validator("type")
    @classmethod
    def validate_type(cls, v):
        if v not in ["K", "C", "M", "T"]:
            raise ValueError(
                'type must be "K" (keyword), "C" (uncaught), '
                '"T" (ticket), or "M" (missed call)'
            )
        return v

    @model_validator(mode="after")
    def validate_match_type(self):
        if self.type == "K" and self.match_type not in ["F", "O", ""]:
            raise ValueError(
                'match_type must be "F" (starts with) or "O" (only) if type is "K".'
            )

        return self



import json
import logging
import os
import shutil
from pathlib import Path

from tablib import Databook, Dataset

from rpft.parsers.universal import UniJSONReader, bookify, parse_tables
from rpft.parsers.creation.contentindexparser import ContentIndexParser
from rpft.parsers.creation.tagmatcher import TagMatcher
from rpft.parsers.sheets import (
    AbstractSheetReader,
    CSVSheetReader,
    GoogleSheetReader,
    JSONSheetReader,
    ODSSheetReader,
    XLSXSheetReader,
)
from rpft.rapidpro.models.containers import RapidProContainer
from rpft.sources import JSONDataSource, SheetDataSource


LOGGER = logging.getLogger(__name__)
FMT_READER_MAP = {
    "csv": CSVSheetReader,
    "google_sheets": GoogleSheetReader,
    "json": JSONSheetReader,
    "ods": ODSSheetReader,
    "uni": UniJSONReader,
    "xlsx": XLSXSheetReader,
}


def create_flows(input_files, output_file, sheet_format, data_models=None, tags=[]):
    """
    Convert source spreadsheet(s) into RapidPro flows.

    :param input_files: list of source spreadsheets to convert
    :param output_files: (deprecated) path of file to export flows to as JSON
    :param sheet_format: format of the spreadsheets
    :param data_models: name of module containing supporting Python data classes
    :param tags: names of tags to be used to filter the source spreadsheets
    :returns: dict representing the RapidPro import/export format.
    """

    try:
        flows = (
            get_content_index_parser(input_files, sheet_format, data_models, tags)
            .parse_all()
            .render()
        )
    except Exception as e:
        LOGGER.critical(e.args[0] if e.args else e.__class__.__name__)
        raise

    if output_file:
        with open(output_file, "w", encoding="utf8") as export:
            json.dump(flows, export, indent=4)

    return flows


def uni_to_sheets(infile) -> bytes:
    with open(infile, "r") as handle:
        data = json.load(handle)

    sheets = bookify(data)
    book = Databook(
        [Dataset(*table[1:], headers=table[0], title=name) for name, table in sheets]
    )

    return book.export("ods")


def sheets_to_uni(infile) -> list:
    return parse_tables(create_sheet_reader(None, infile))


def get_content_index_parser(input_files, sheet_format, data_models, tags):
    if not sheet_format and not data_models:
        return ContentIndexParser(
            JSONDataSource(input_files), data_models, TagMatcher(tags)
        )

    readers = [
        create_sheet_reader(sheet_format, input_file) for input_file in input_files
    ]

    return ContentIndexParser(SheetDataSource(readers), data_models, TagMatcher(tags))


def convert_to_json(input_file, sheet_format):
    """
    Convert source spreadsheet(s) into json.

    :param input_file: source spreadsheet to convert
    :param sheet_format: format of the input spreadsheet
    :returns: content of the input file converted to json.
    """

    return to_json(create_sheet_reader(sheet_format, input_file))


def flows_to_sheets(
    input_file, output_folder, format="csv", strip_uuids=False, numbered=False
):
    """
    Convert source RapidPro JSON to spreadsheet(s).

    Each flow in the JSON will become a separate output file.

    :param input_file: source JSON file to convert
    :param output_folder: destination folder for output files
    :param format: Output file format.
    :param strip_uuids: Strip all UUIDs from output to allow for comparing outputs.
    :param numbered: Use sequential numbers instead of short reps for row IDs.
    :returns: None.
    """
    with open(input_file, "r", encoding="utf-8") as f:
        data = json.load(f)
    container = RapidProContainer.from_dict(data)
    for flow in container.flows:
        rds = flow.to_row_data_sheet(strip_uuids, numbered)
        rds.export(os.path.join(output_folder, f"{flow.name}.{format}"), format)


def create_sheet_reader(sheet_format, input_file):
    cls = FMT_READER_MAP.get(sheet_format) or next(
        reader for reader in FMT_READER_MAP.values() if reader.can_process(input_file)
    )

    if cls:
        return cls(input_file)

    raise Exception(f"Format not supported, file={input_file}")


def sheets_to_csv(path, sheet_ids):
    prepare_dir(path)

    for sid in sheet_ids:
        sheet_to_csv(path, sid)


def sheet_to_csv(path, sheet_id):
    workbook_dir = prepare_dir(Path(path) / sheet_id)
    reader = GoogleSheetReader(sheet_id)

    for name, sheet in reader.sheets.items():
        with open(
            workbook_dir / f"{name}.csv",
            "w",
            newline="",
            encoding="utf-8",
        ) as csv_file:
            csv_file.write(sheet.table.export("csv"))


def to_json(reader: AbstractSheetReader) -> str:
    book = {
        "meta": {
            "version": "0.1.0",
        },
        "sheets": {name: sheet.table.dict for name, sheet in reader.sheets.items()},
    }

    return json.dumps(book, ensure_ascii=False, indent=2)


def prepare_dir(path):
    directory = Path(path)

    if directory.exists():
        shutil.rmtree(directory)
    directory.mkdir(parents=True)

    return directory



import unittest

from rpft.parsers.common.cellparser import CellParser
from rpft.parsers.common.rowdatasheet import RowDataSheet
from rpft.parsers.common.rowparser import RowParser
from rpft.parsers.creation.flowrowmodel import (
    Edge,
    FlowRowModel,
    Webhook,
    list_of_pairs_to_dict,
)
from rpft.rapidpro.models.actions import (
    AddContactGroupAction,
    Group,
    SendMessageAction,
    SetContactFieldAction,
    SetRunResultAction,
    WhatsAppMessageTemplating,
)
from rpft.rapidpro.models.common import mangle_string
from rpft.rapidpro.models.containers import FlowContainer
from rpft.rapidpro.models.nodes import (
    BasicNode,
    CallWebhookNode,
    EnterFlowNode,
    RandomRouterNode,
    SwitchRouterNode,
)
from tests.row_data import (
    get_message_with_templating,
    get_start_row,
    get_unconditional_node_from_1,
)


class TestMangle(unittest.TestCase):
    def test_mangle(self):
        mangled = mangle_string("ab@cd ef_gh-ij1234567")
        self.assertEqual(mangled, "abcd_ef_gh-ij12")


class TestToRowModels(unittest.TestCase):
    def compare_row_models_without_uuid(self, row_models1, row_models2):
        self.maxDiff = None
        self.assertEqual(len(row_models1), len(row_models2))
        for model1, model2 in zip(row_models1, row_models2):
            data1 = model1.model_dump()
            data2 = model2.model_dump()
            if not data1["node_uuid"] or not data2["node_uuid"]:
                # If one of them is blank, skip the comparison
                data1.pop("node_uuid")
                data2.pop("node_uuid")
            self.assertEqual(data1, data2)


class TestNodes(TestToRowModels):
    def test_basic_node(self):
        row_data = get_start_row()
        node = BasicNode()
        action = SendMessageAction(
            row_data.mainarg_message_text, quick_replies=row_data.choices
        )
        node.add_action(action)
        node.initiate_row_models(1, Edge(from_="start"))
        row_models = node.get_row_models()
        self.compare_row_models_without_uuid(row_models, [row_data])

    def test_templating_node(self):
        row_data = get_message_with_templating()
        templating = WhatsAppMessageTemplating.from_whats_app_templating_model(
            row_data.wa_template
        )
        node = BasicNode()
        action = SendMessageAction(
            row_data.mainarg_message_text,
            quick_replies=row_data.choices,
            templating=templating,
        )
        node.add_action(action)
        node.initiate_row_models(1, Edge(from_="start"))
        row_models = node.get_row_models()
        self.compare_row_models_without_uuid(row_models, [row_data])

    def test_basic_node_two_actions(self):
        row_data1 = get_start_row()
        row_data2 = get_unconditional_node_from_1()
        node = BasicNode()
        action = SendMessageAction(
            row_data1.mainarg_message_text, quick_replies=row_data1.choices
        )
        node.add_action(action)
        action = SendMessageAction(
            row_data2.mainarg_message_text, quick_replies=row_data2.choices
        )
        node.add_action(action)
        node.initiate_row_models("1", Edge(from_="start"))
        row_models = node.get_row_models()
        # The section action gets row ID 1.1, which needs to be remapped to 2.
        # When converting flows to rows, this remapping is done automatically.
        row_models[1].row_id = "2"
        self.compare_row_models_without_uuid(row_models, [row_data1, row_data2])

    def test_add_group_node(self):
        row_data = FlowRowModel(
            **{
                "row_id": "1",
                "type": "add_to_group",
                "mainarg_groups": ["test group"],
                "obj_id": "8224bfe2-acec-434f-bc7c-14c584fc4bc8",
                "edges": [{"from_": "start"}],
                "ui_position": ["123", "456"],
                "node_uuid": "224f6caa-fd25-47d3-96a9-3d43506b7878",
            }
        )
        node = BasicNode(
            uuid=row_data.node_uuid, ui_pos=[int(p) for p in row_data.ui_position]
        )
        action = AddContactGroupAction(
            [Group(row_data.mainarg_groups[0], row_data.obj_id)]
        )
        node.add_action(action)
        node.initiate_row_models(1, Edge(from_="start"))
        row_models = node.get_row_models()
        self.compare_row_models_without_uuid(row_models, [row_data])


class TestFlowContainer(TestToRowModels):
    def test_basic_node(self):
        row_data = get_start_row()
        node = BasicNode()
        action = SendMessageAction(
            row_data.mainarg_message_text, quick_replies=row_data.choices
        )
        node.add_action(action)
        container = FlowContainer("test_flow")
        container.add_node(node)
        row_models = container.to_rows(numbered=True)
        self.compare_row_models_without_uuid(row_models, [row_data])

    def test_webhook_node(self):
        webhook_data = {
            "url": "the_url",
            "method": "POST",
            "body": "payload",
        }
        headers = [["header1", "value1"], ["header2", "value2"]]
        headers_dict = list_of_pairs_to_dict(headers)
        row_data = FlowRowModel(
            row_id="webhook.the_url",
            edges=[{"from_": "start"}],
            type="call_webhook",
            webhook=Webhook(headers=headers, **webhook_data),
            save_name="result name",
        )
        node = CallWebhookNode(
            result_name="result name", headers=headers_dict, **webhook_data
        )

        container = FlowContainer("test_flow")
        container.add_node(node)
        row_models = container.to_rows()
        self.compare_row_models_without_uuid(row_models, [row_data])

    def test_two_basic_nodes(self):
        row_data1 = get_start_row()
        row_data2 = get_unconditional_node_from_1()
        node1 = BasicNode()
        node2 = BasicNode()
        action1 = SendMessageAction(
            row_data1.mainarg_message_text, quick_replies=row_data1.choices
        )
        action2 = SendMessageAction(
            row_data2.mainarg_message_text, quick_replies=row_data2.choices
        )
        node1.add_action(action1)
        node2.add_action(action2)
        node1.update_default_exit(node2.uuid)
        container = FlowContainer("test_flow")
        container.add_node(node1)
        container.add_node(node2)
        row_models = container.to_rows(numbered=True)
        self.compare_row_models_without_uuid(row_models, [row_data1, row_data2])

    def test_conditional_edge(self):
        row_data1 = FlowRowModel(
            **{
                "row_id": "switch.fields_name",
                "type": "split_by_value",
                "edges": [{"from_": "start"}],
                "mainarg_expression": "@fields.name",
            }
        )
        row_data2 = FlowRowModel(
            **{
                "row_id": "msg.Message_if_fiel",
                "type": "send_message",
                "edges": [
                    {
                        "from_": "switch.fields_name",
                        "condition": {
                            "value": "3",
                            "variable": "@fields.name",
                            "type": "has_phrase",
                            "name": "3",
                        },
                    }
                ],
                "mainarg_message_text": "Message if @fields.name == 3",
            }
        )
        case = row_data2.edges[0].condition
        container = FlowContainer("test_flow")

        node1 = SwitchRouterNode(case.variable)
        node2 = BasicNode()

        node1.add_choice(case.variable, case.type, [case.value], case.name, node2.uuid)
        container.add_node(node1)

        action2 = SendMessageAction(
            row_data2.mainarg_message_text, quick_replies=row_data2.choices
        )
        node2.add_action(action2)
        container.add_node(node2)

        row_models = container.to_rows()
        self.compare_row_models_without_uuid(row_models, [row_data1, row_data2])

    def test_conditional_edge2(self):
        row_data1 = FlowRowModel(
            **{
                "row_id": "switch.contact_groups",
                "type": "split_by_group",
                "edges": [{"from_": "start"}],
                "mainarg_groups": ["my group"],
                "obj_id": "12345678",
            }
        )
        row_data2 = FlowRowModel(
            **{
                "row_id": "set_contact_field.my_variable",
                "type": "save_value",
                "edges": [
                    {
                        "from_": "switch.contact_groups",
                        "condition": {"value": "my group"},
                    }
                ],
                "mainarg_value": "my value",
                "save_name": "my_variable",
            }
        )
        case = row_data2.edges[0].condition
        container = FlowContainer("test_flow")

        node1 = SwitchRouterNode("@contact.groups")
        node2 = BasicNode()

        # The arguments for this choice is [group uuid, group name]
        node1.add_choice(
            case.variable,
            case.type or "has_group",
            [row_data1.obj_id, case.value],
            case.name,
            node2.uuid,
        )
        container.add_node(node1)

        action2 = SetContactFieldAction(row_data2.save_name, row_data2.mainarg_value)
        node2.add_action(action2)
        container.add_node(node2)

        row_models = container.to_rows()
        self.compare_row_models_without_uuid(row_models, [row_data1, row_data2])

    def test_random_double_edge(self):
        # Two edges into the same node (i.e. forward edge in the tree)
        # Uses random router

        row_data1 = FlowRowModel(
            **{
                "row_id": "random",
                "type": "split_random",
                "edges": [{"from_": "start"}],
            }
        )

        row_data2 = FlowRowModel(
            **{
                "row_id": "msg.Second_node_mes",
                "type": "send_message",
                "edges": [
                    {
                        "from_": "random",
                        "condition": {"value": "1"},
                    },
                    {
                        "from_": "random",
                        "condition": {"value": "2"},
                    },
                ],
                "mainarg_message_text": "Second node message",
            }
        )

        container = FlowContainer("test_flow")

        node1 = RandomRouterNode()
        node2 = BasicNode()

        node1.add_choice(row_data2.edges[0].condition.value, node2.uuid)
        node1.add_choice(row_data2.edges[1].condition.value, node2.uuid)
        container.add_node(node1)

        action2 = SendMessageAction(row_data2.mainarg_message_text)
        node2.add_action(action2)
        container.add_node(node2)

        row_models = container.to_rows()
        self.compare_row_models_without_uuid(row_models, [row_data1, row_data2])

    def test_cyclic_wait_edge(self):
        # An edge for a wait_for_response node to itself (cycle) via go_to

        row_data1 = FlowRowModel(
            **{
                "row_id": "wait_for.wait_result",
                "type": "wait_for_response",
                "save_name": "wait_result",
                "no_response": "300",
                "edges": [{"from_": "start"}],
            }
        )
        # A proper case
        row_data2 = FlowRowModel(
            **{
                "row_id": "goto.wait_for.wait_result",
                "type": "go_to",
                "edges": [
                    {
                        "from_": "wait_for.wait_result",
                        "condition": {
                            "value": "word",
                            "variable": "@input.text",
                            "type": "has_any_word",
                            "name": "Word",
                        },
                    }
                ],
                "mainarg_destination_row_ids": ["wait_for.wait_result"],
            }
        )
        # The 'Other' category
        row_data3 = FlowRowModel(
            **{
                "row_id": "msg.Node_message_te",
                "type": "send_message",
                "edges": [
                    {
                        "from_": "wait_for.wait_result",
                    }
                ],
                "mainarg_message_text": "Node message text number 2",
            }
        )
        # The no response category
        row_data4 = FlowRowModel(
            **{
                "row_id": "msg.Node_message_te.1",
                "type": "send_message",
                "edges": [
                    {
                        "from_": "wait_for.wait_result",
                        "condition": {"value": "No Response"},
                    }
                ],
                "mainarg_message_text": "Node message text number 3",
            }
        )

        case1 = row_data2.edges[0].condition
        container = FlowContainer("test_flow")
        node1 = SwitchRouterNode(
            "@input.text",
            result_name=row_data1.save_name,
            wait_timeout=int(row_data1.no_response),
        )
        node2 = BasicNode()
        node3 = BasicNode()
        node1.add_choice(
            case1.variable, case1.type, [case1.value], case1.name, node1.uuid
        )
        node1.update_default_exit(node2.uuid)
        node1.update_no_response_exit(node3.uuid)
        container.add_node(node1)

        action2 = SendMessageAction(row_data3.mainarg_message_text)
        node2.add_action(action2)
        container.add_node(node2)
        action3 = SendMessageAction(row_data4.mainarg_message_text)
        node3.add_action(action3)
        container.add_node(node3)

        row_models = container.to_rows()
        self.compare_row_models_without_uuid(
            row_models, [row_data1, row_data2, row_data3, row_data4]
        )

    def test_enter_flow_edge(self):
        # Test expired (default) edge of enter_flow node
        # test save_flow_result with a @result (set_run_result in rapidpro)

        row_data1 = FlowRowModel(
            **{
                "row_id": "flow.sample_flow",
                "type": "start_new_flow",
                "mainarg_flow_name": "sample_flow",
                "obj_id": "8224bfe2-acec-434f-bc7c-14c584fc4bc8",
                "edges": [{"from_": "start"}],
            }
        )

        row_data2 = FlowRowModel(
            **{
                "row_id": "set_run_result.my_result",
                "type": "save_flow_result",
                "edges": [
                    {
                        "from_": "flow.sample_flow",
                        "condition": {"value": "expired"},
                    }
                ],
                "mainarg_value": "Value",
                "save_name": "my result",
            }
        )

        case1 = row_data2.edges[0].condition
        container = FlowContainer("test_flow")

        node1 = EnterFlowNode(row_data1.mainarg_flow_name, row_data1.obj_id)
        node2 = BasicNode()
        node1.add_choice(
            case1.variable,
            case1.type or "has_only_text",
            [case1.value],
            case1.name,
            node2.uuid,
            is_default=True,
        )
        container.add_node(node1)

        action2 = SetRunResultAction(row_data2.save_name, row_data2.mainarg_value)
        node2.add_action(action2)
        container.add_node(node2)

        row_models = container.to_rows()
        self.compare_row_models_without_uuid(row_models, [row_data1, row_data2])


class TestRowModelExport(unittest.TestCase):
    def test_row_model_export(self):
        row_data = FlowRowModel(
            **{
                "row_id": "1",
                "type": "send_message",
                "mainarg_message_text": "Hello",
                "choices": ["QR1", "QR2", "QR3"],
                "edges": [{"from_": "start"}],
                "node_uuid": "224f6caa-fd25-47d3-96a9-3d43506b7878",
                "ui_position": ["123", "456"],
            }
        )

        # Because default values are '', even blank fields are exported.
        # This may change in the future.
        expected_headers = [
            "row_id",
            "type",
            "edges.1.from",
            "message_text",
            "choices.1",
            "choices.2",
            "choices.3",
            "_nodeId",
            "_ui_position",
        ]
        expected_content = (
            "1",
            "send_message",
            "start",
            "Hello",
            "QR1",
            "QR2",
            "QR3",
            "224f6caa-fd25-47d3-96a9-3d43506b7878",
            "123|456",
        )

        sheet = RowDataSheet(RowParser(FlowRowModel, CellParser()), [row_data])
        tablib_sheet = sheet.convert_to_tablib()
        self.assertEqual(tablib_sheet.headers, expected_headers)
        self.assertEqual(tablib_sheet[0], expected_content)



import copy

import tablib

from rpft.parsers.common.sheetparser import SheetParser
from rpft.parsers.sheets import AbstractSheetReader, Sheet


class MockCellParser:
    def parse(self, value, context={}):
        return value

    def parse_as_string(self, value, context={}):
        return value, False

    def join_from_lists(self, nested_list):
        return nested_list


class MockRowParser:
    def __init__(self):
        self.context = {}

    def unparse_row(self, row, target_headers=set(), excluded_headers=set()):
        return row

    def parse_row(self, row, template_context):
        row = copy.deepcopy(row)
        row["context"] = template_context
        return row


class MockSheetParser(SheetParser):
    def __init__(self, rows, context={}):
        """
        Args:
            rows: List of instances of the RowModel
            context: context used for template parsing
        """

        self.bookmarks = {}
        self.input_rows = rows
        self.iterator = iter(self.input_rows)
        self.context = copy.deepcopy(context)

    def parse_next_row(self, omit_templating=False, return_index=False):
        try:
            input_row = next(self.iterator)
        except StopIteration:
            return (None, None) if return_index else None
        return (input_row, -1) if return_index else None

    def get_row_data_sheet(self):
        raise NotImplementedError


class MockSheetReader(AbstractSheetReader):
    def __init__(self, main_sheet_data=None, sheet_data_dict={}, name="mock"):
        self.name = name
        self._sheets = {}

        if main_sheet_data:
            self._sheets["content_index"] = Sheet(
                reader=self,
                name="content_index",
                table=tablib.import_set(main_sheet_data, format="csv"),
            )

        for name, content in sheet_data_dict.items():
            self._sheets[name] = Sheet(
                reader=self,
                name=name,
                table=tablib.import_set(content, format="csv"),
            )



from rpft.parsers.common.rowparser import ParserModel


class Condition(ParserModel):
    value: str = ""
    """
    Value to check against. (Does not support checks with a list of values).
    """

    variable: str = ""
    """
    Variable or expression that is being checked against value.
    """

    type: str = ""
    """
    Type of the conditional check, e.g. has_any_word (default).
    """

    name: str = ""
    """
    Display name in RapidPro only.
    """


class ConditionWithMessage(ParserModel):
    condition: Condition
    message: str


class ConditionWithId(ParserModel):
    condition: Condition = Condition()
    rowid: str = ""


class ConditionsWithMessage(ParserModel):
    conditions: list[ConditionWithMessage] = []
    general_message: str = ""
    external_flow: str = ""


class Assignment(ParserModel):
    """
    Assign a value to a variable
    """

    variable: str
    value: str


class Expiration(ParserModel):
    time: str = ""
    """
    Time after which a flow expires.
    """

    message: str = ""
    """
    Message to send when the flow expires.
    """


class Message(ParserModel):
    text: str
    image: str = ""
    audio: str = ""
    video: str = ""
    attachments: list[str] = []


class TemplateSheet:
    def __init__(self, name, table, argument_definitions):
        self.name = name
        self.table = table
        self.argument_definitions = argument_definitions


class ChatbotDefinition:
    def __init__(
        self,
        flow_definitions,
        data_sheets,
        templates: list[TemplateSheet],
        surveys,
        survey_questions,
        global_context=None,
    ):
        self.flow_definitions = flow_definitions
        self.data_sheets = data_sheets
        self.templates = templates
        self.surveys = surveys
        self.survey_questions = survey_questions
        self.global_context = global_context or {}

    def get_data_sheet_rows(self, sheet_name):
        return self.data_sheets[sheet_name].rows

    def get_data_sheet_row(self, sheet_name, row_id):
        return self.data_sheets[sheet_name].rows[row_id]

    def get_template(self, name) -> TemplateSheet:
        return self.templates[name]


class SurveyConfig(ParserModel):
    variable_prefix: str = ""
    """
    Prefix to apply to all variable names in the survey.
    """

    expiration_message: str = ""
    """
    Message to send when a question flow expires.
    """


class MCQChoice(ParserModel):
    text: str
    """
    Text rendered to user.
    """

    value: str
    """
    Text stored in survey variable.
    """

    show: str = ""
    """
    Whether to display the choice as a quick reply. A value of "no" won't display the
    choice as quick reply, but will add the case to the WFR node.
    """
    response: str = ""
    """
    Response message send to the user if they select that choice before moving to the next question
    """

class PostProcessing(ParserModel):
    assignments: list[Assignment] = []
    """
    Assignments to perform via save_value rows.
    """

    flow: str = ""
    """
    Flow to invoke for postprocessing.
    """


class SkipOption(ParserModel):
    text: str = ""
    """
    The text with instructions that the user sees.
    """

    choice: str = ""
    """
    The choice (quick reply/text) the user has to enter to skip this question.
    """

    value: str = ""
    """
    The value that is stored in the question variable if the question is skipped.
    """


class Confirmation(ParserModel):
    condition: Condition = Condition()
    question: str = ""
    confirm_option: str = ""
    back_option: str = ""


class SurveyQuestionModel(ParserModel):
    """
    Representation of a survey question.
    """

    type: str
    """
    Type of the question.
    """

    messages: list[Message]
    """
    Question text.
    """

    variable: str = ""
    """
    Variable to store the user input in. If blank, generated from the question ID as
    sq_{survey_id}_{question_id}. The survey_id/question_id is the survey's
    name/question's ID in all lowercase with non-alphanumeric characters removed.
    """

    completion_variable: str = ""
    """
    Variable indicating whether question has been completed. If blank, generated from
    the question ID as {variable}_complete.
    """

    choices: list[MCQChoice] = []
    """
    MCQ specific fields.
    """

    expiration: Expiration = Expiration()
    """
    Message to send when question flow expires. If blank, message from survey
    configuration is used.
    """

    relevant: list[Condition] = []
    """
    Conditions required to present the question, otherwise skipped.
    """

    skipoption: SkipOption = SkipOption()
    """
    Add an additional choice to allow the question to be skipped.
    """

    confirmation: Confirmation = Confirmation()
    """
    Seek confirmation of the answer if a condition is met. The choices will be 'Yes'
    (store the answer and proceed) or 'No' (repeat the question).
    """

    stop: ConditionsWithMessage = ConditionsWithMessage()
    """
    End the survey if any of the conditions are met.
    """

    back: ConditionWithId = ConditionWithId()
    """
    Return to the previous question in the survey if the condition is met.
    """

    validation: ConditionsWithMessage = ConditionsWithMessage()
    """
    Repeat the question of any of the conditions are met.
    """

    postprocessing: PostProcessing = PostProcessing()
    """
    Variable postprocessing. Postprocessing to do after a user's answer is successfully
    stored. This could be an assignment (of the same or another variable), or a flow
    that is triggered.
    """

    tags: list[str] = []
    """
    Tags allowing to filter questions to appear in a survey.
    """

    def header_name_to_field_name_with_context(header, row):
        if header == "question":
            return "messages.1.text"
        elif header == "image":
            return "messages.1.image"
        else:
            return header



DEFAULT_CONFIG = {
    "disable_existing_loggers": False,
    "filters": {
        "context": {
            "()": "rpft.logger.logger.ContextFilter",
        }
    },
    "formatters": {
        "default": {
            "format": "%(levelname)s:%(name)s: %(processing_stack)s: %(message)s",
        },
    },
    "handlers": {
        "file": {
            "class": "logging.FileHandler",
            "filename": "errors.log",
            "filters": ["context"],
            "formatter": "default",
            "level": "INFO",
            "mode": "w",
        },
        "console": {
            "class": "logging.StreamHandler",
            "filters": ["context"],
            "formatter": "default",
            "level": "CRITICAL",
        },
    },
    "root": {
        "level": "INFO",
        "handlers": [
            "file",
            "console",
        ],
    },
    "version": 1,
}



from typing import Set
from unittest import TestCase

from rpft.parsers.creation.contentindexparser import ContentIndexParser
from rpft.parsers.creation.tagmatcher import TagMatcher
from rpft.parsers.sheets import (
    CSVSheetReader,
    DatasetSheetReader,
    XLSXSheetReader,
)
from rpft.rapidpro.models.triggers import RapidProTriggerError
from rpft.rapidpro.simulation import Context, traverse_flow
from rpft.sources import SheetDataSource

from tablib import Dataset
from tests import TESTS_ROOT
from tests.mocks import MockSheetReader
from tests.utils import csv_join


class TestTemplate(TestCase):
    def assertFlowMessages(self, render_output, flow_name, messages_exp, context=None):
        flows = [flow for flow in render_output["flows"] if flow["name"] == flow_name]

        self.assertTrue(
            len(flows) > 0,
            msg=f'Flow with name "{flow_name}" does not exist in output.',
        )

        actions = traverse_flow(flows[0], context or Context())
        actions_exp = list(zip(["send_msg"] * len(messages_exp), messages_exp))

        self.assertEqual(actions, actions_exp)


class TestTemplateDefinition(TestCase):
    def setUp(self):
        template = csv_join(
            "row_id,type,from,message_text",
            ",send_message,start,Some text",
        )
        self.sheet_data_dict = {
            "my_template": template,
            "my_template2": template,
        }

    def test_basic_template_definition(self):
        ci_sheet = (
            "type,sheet_name,data_sheet,data_row_id,new_name,data_model,status\n"
            "template_definition,my_template,,,,,\n"
            "template_definition,my_template2,,,,,draft\n"
        )
        self.parser = ContentIndexParser(
            SheetDataSource(
                [
                    MockSheetReader(
                        ci_sheet,
                        self.sheet_data_dict,
                    )
                ]
            )
        )

        self.assertTemplateDefinition()

    def test_ignore_template_definition(self):
        """
        Ensure that ignoring a template row does NOT remove the template
        """
        ci_sheet = (
            "type,sheet_name,data_sheet,data_row_id,new_name,data_model,status\n"
            "template_definition,my_template,,,,,\n"
            "ignore_row,my_template,,,,,\n"
        )
        self.parser = ContentIndexParser(
            SheetDataSource(
                [
                    MockSheetReader(
                        ci_sheet,
                        self.sheet_data_dict,
                    )
                ]
            )
        )

        self.assertTemplateDefinition()

    def assertTemplateDefinition(self):
        template_sheet = self.parser.template_sheets["my_template"]

        self.assertEqual(template_sheet.table[0][1], "send_message")
        self.assertEqual(template_sheet.table[0][3], "Some text")
        with self.assertRaises(KeyError):
            self.parser.template_sheets["my_template2"]


class TestParsing(TestTemplate):
    def test_basic_nesting(self):
        ci_sheet = (
            "type,sheet_name,data_sheet,data_row_id,new_name,data_model,status\n"
            "template_definition,my_template,,,,,\n"
            "content_index,ci_sheet2,,,,,\n"
        )
        ci_sheet2 = (
            "type,sheet_name,data_sheet,data_row_id,new_name,data_model,status\n"
            "template_definition,my_template2,,,,,\n"
        )
        my_template = csv_join(
            "row_id,type,from,message_text",
            ",send_message,start,Some text",
        )
        my_template2 = csv_join(
            "row_id,type,from,message_text",
            ",send_message,start,Other text",
        )
        sheet_dict = {
            "ci_sheet2": ci_sheet2,
            "my_template": my_template,
            "my_template2": my_template2,
        }
        ci_parser = ContentIndexParser(
            SheetDataSource([MockSheetReader(ci_sheet, sheet_dict)])
        )

        self.assertEqual(
            ci_parser.template_sheets["my_template"].table[0][3],
            "Some text",
        )
        self.assertEqual(
            ci_parser.template_sheets["my_template2"].table[0][3],
            "Other text",
        )

    def test_basic_user_model(self):
        ci_sheet = (
            "type,sheet_name,data_sheet,data_row_id,new_name,data_model,status\n"
            "data_sheet,simpledata,,,,SimpleRowModel,\n"
        )
        simpledata = csv_join(
            "ID,value1,value2",
            "rowA,1A,2A",
            "rowB,1B,2B",
        )
        definition = ContentIndexParser(
            SheetDataSource([MockSheetReader(ci_sheet, {"simpledata": simpledata})]),
            "tests.datarowmodels.simplemodel",
        ).definition
        datamodelA = definition.get_data_sheet_row("simpledata", "rowA")
        datamodelB = definition.get_data_sheet_row("simpledata", "rowB")

        self.assertEqual(datamodelA.value1, "1A")
        self.assertEqual(datamodelA.value2, "2A")
        self.assertEqual(datamodelB.value1, "1B")
        self.assertEqual(datamodelB.value2, "2B")

    def test_flow_type(self):
        ci_sheet = (
            "type,sheet_name,new_name,data_model,options\n"
            "create_flow,my_basic_flow,,,\n"
            "create_flow,my_basic_flow,my_other_flow,,flow_type;messaging_background\n"
        )
        my_basic_flow = csv_join(
            "row_id,type,from,message_text",
            ",send_message,start,Some text",
        )
        sheet_dict = {
            "my_basic_flow": my_basic_flow,
        }
        render_output = (
            ContentIndexParser(
                SheetDataSource([MockSheetReader(ci_sheet, sheet_dict)]),
                "tests.datarowmodels.nestedmodel",
            )
            .parse_all()
            .render()
        )

        self.assertEqual(len(render_output["flows"]), 2)
        self.assertEqual(render_output["flows"][0]["name"], "my_basic_flow")
        self.assertEqual(render_output["flows"][0]["type"], "messaging")
        self.assertEqual(render_output["flows"][1]["name"], "my_other_flow")
        self.assertEqual(render_output["flows"][1]["type"], "messaging_background")

    def test_ignore_flow_definition(self):
        ci_sheet = (
            "type,sheet_name,data_sheet,data_row_id,new_name,data_model,status\n"
            "create_flow,my_basic_flow,,,,,\n"
            "create_flow,my_basic_flow,,,my_renamed_basic_flow,,\n"
            "create_flow,my_basic_flow,,,my_renamed_basic_flow2,,\n"
            "ignore_row,my_basic_flow,,,,,\n"
            "ignore_row,my_renamed_basic_flow,,,,,\n"
        )
        my_basic_flow = csv_join(
            "row_id,type,from,message_text",
            ",send_message,start,Some text",
        )
        sheet_dict = {
            "my_basic_flow": my_basic_flow,
        }
        render_output = (
            ContentIndexParser(
                SheetDataSource([MockSheetReader(ci_sheet, sheet_dict)]),
                "tests.datarowmodels.nestedmodel",
            )
            .parse_all()
            .render()
        )

        self.assertEqual(len(render_output["flows"]), 1)
        self.assertEqual(render_output["flows"][0]["name"], "my_renamed_basic_flow2")

    def test_ignore_templated_flow_definition(self):
        ci_sheet = (
            "type,sheet_name,data_sheet,data_row_id,new_name,data_model,status\n"
            "create_flow,my_template,nesteddata,,,,\n"
            "create_flow,my_template,nesteddata,,bulk_renamed,,\n"
            "create_flow,my_template,nesteddata,row1,row1_renamed,,\n"
            "create_flow,my_template,nesteddata,row2,row2_renamed,,\n"
            "create_flow,my_template,nesteddata,row2,,,\n"
            "ignore_row,my_template,,,,,\n"
            "ignore_row,row1_renamed,,,,,\n"
            "data_sheet,nesteddata,,,,NestedRowModel,\n"
        )
        nesteddata = (
            "ID,value1,custom_field.happy,custom_field.sad\n"
            "row1,Value1,Happy1,Sad1\n"
            "row2,Value2,Happy2,Sad2\n"
        )
        my_template = (
            "row_id,type,from,message_text\n"
            ",send_message,start,{{value1}}\n"
            ",send_message,,{{custom_field.happy}} and {{custom_field.sad}}\n"
        )
        sheet_dict = {
            "nesteddata": nesteddata,
            "my_template": my_template,
        }
        render_output = (
            ContentIndexParser(
                SheetDataSource([MockSheetReader(ci_sheet, sheet_dict)]),
                "tests.datarowmodels.nestedmodel",
            )
            .parse_all()
            .render()
        )

        self.assertEqual(len(render_output["flows"]), 3)
        self.assertFlowNamesEqual(
            render_output,
            {
                "bulk_renamed - row1",
                "bulk_renamed - row2",
                "row2_renamed - row2",
            },
        )

    def test_generate_flows(self):
        ci_sheet = (
            "type,sheet_name,data_sheet,data_row_id,new_name,data_model,status\n"
            "create_flow,my_template,nesteddata,row1,,,\n"
            "create_flow,my_template,nesteddata,row2,,,\n"
            "create_flow,my_basic_flow,,,,,\n"
            "data_sheet,nesteddata,,,,NestedRowModel,\n"
        )
        ci_sheet_alt = (
            "type,sheet_name,data_sheet,data_row_id,new_name,data_model,status\n"
            "create_flow,my_template,nesteddata,,,,\n"
            "create_flow,my_basic_flow,,,,,\n"
            "data_sheet,nesteddata,,,,NestedRowModel,\n"
        )
        nesteddata = (
            "ID,value1,custom_field.happy,custom_field.sad\n"
            "row1,Value1,Happy1,Sad1\n"
            "row2,Value2,Happy2,Sad2\n"
            "row3,Value3,Happy3,Sad3\n"
        )
        my_template = (
            "row_id,type,from,message_text\n"
            ",send_message,start,{{value1}}\n"
            ",send_message,,{{custom_field.happy}} and {{custom_field.sad}}\n"
        )
        my_basic_flow = csv_join(
            "row_id,type,from,message_text",
            ",send_message,start,Some text",
        )
        sheet_dict = {
            "nesteddata": nesteddata,
            "my_template": my_template,
            "my_basic_flow": my_basic_flow,
        }
        render_output = (
            ContentIndexParser(
                SheetDataSource([MockSheetReader(ci_sheet, sheet_dict)]),
                "tests.datarowmodels.nestedmodel",
            )
            .parse_all()
            .render()
        )

        self.assertFlowMessages(
            render_output,
            "my_basic_flow",
            ["Some text"],
        )
        self.assertFlowMessages(
            render_output,
            "my_template - row1",
            ["Value1", "Happy1 and Sad1"],
        )
        self.assertFlowMessages(
            render_output,
            "my_template - row2",
            ["Value2", "Happy2 and Sad2"],
        )

        render_output = (
            ContentIndexParser(
                SheetDataSource([MockSheetReader(ci_sheet_alt, sheet_dict)]),
                "tests.datarowmodels.nestedmodel",
            )
            .parse_all()
            .render()
        )

        self.assertFlowMessages(
            render_output,
            "my_basic_flow",
            ["Some text"],
        )
        self.assertFlowMessages(
            render_output,
            "my_template - row1",
            ["Value1", "Happy1 and Sad1"],
        )
        self.assertFlowMessages(
            render_output,
            "my_template - row2",
            ["Value2", "Happy2 and Sad2"],
        )
        self.assertFlowMessages(
            render_output,
            "my_template - row3",
            ["Value3", "Happy3 and Sad3"],
        )

    def test_duplicate_create_flow(self):
        ci_sheet = (
            "type,sheet_name,data_sheet,data_row_id,new_name,data_model,status\n"
            "create_flow,my_template,,,,,\n"
            "create_flow,my_template2,,,my_template,,\n"
        )
        my_template = csv_join(
            "row_id,type,from,message_text",
            ",send_message,start,Some text",
        )
        my_template2 = csv_join(
            "row_id,type,from,message_text",
            ",send_message,start,Other text",
        )
        sheet_dict = {
            "ci_sheet": ci_sheet,
            "my_template": my_template,
            "my_template2": my_template2,
        }
        render_output = (
            ContentIndexParser(SheetDataSource([MockSheetReader(ci_sheet, sheet_dict)]))
            .parse_all()
            .render()
        )

        self.assertFlowMessages(render_output, "my_template", ["Other text"])

    def test_bulk_flows_with_args(self):
        ci_sheet = (
            "type,sheet_name,data_sheet,data_row_id,template_arguments,new_name,data_model,status\n"  # noqa: E501
            "template_definition,my_template,,,arg1;arg2,,,\n"
            "create_flow,my_template,nesteddata,,ARG1;ARG2,my_renamed_template,,\n"
            "data_sheet,nesteddata,,,,,NestedRowModel,\n"
        )
        nesteddata = (
            "ID,value1,custom_field.happy,custom_field.sad\n"
            "row1,Value1,Happy1,Sad1\n"
            "row2,Value2,Happy2,Sad2\n"
        )
        my_template = (
            "row_id,type,from,message_text\n"
            ",send_message,start,{{value1}} {{arg1}} {{arg2}}\n"
            ",send_message,,{{custom_field.happy}} and {{custom_field.sad}}\n"
        )
        sheet_dict = {
            "nesteddata": nesteddata,
            "my_template": my_template,
        }
        render_output = (
            ContentIndexParser(
                SheetDataSource([MockSheetReader(ci_sheet, sheet_dict)]),
                "tests.datarowmodels.nestedmodel",
            )
            .parse_all()
            .render()
        )

        self.assertFlowMessages(
            render_output,
            "my_renamed_template - row1",
            ["Value1 ARG1 ARG2", "Happy1 and Sad1"],
        )
        self.assertFlowMessages(
            render_output,
            "my_renamed_template - row2",
            ["Value2 ARG1 ARG2", "Happy2 and Sad2"],
        )

    def test_global_context(self):
        ci_sheet = (
            "type,sheet_name,data_sheet,data_row_id,template_arguments,new_name,data_model,status\n"  # noqa: E501
            "create_flow,my_basic_flow,,,,,,\n"
            "globals,globals_sheet1,,,,,,\n"
            "globals,globals_sheet2,,,,,,\n"
            "data_sheet,minimaldata,,,,,NameModel,\n"
            "template_definition,my_template,,,message,,,\n"
            "create_flow,my_template,minimaldata,,Local message,,,\n"
        )
        my_basic_flow = csv_join(
            "row_id,type,from,message_text",
            ",send_message,start,{{globals.message}}",
            ",send_message,,{{globals['spaced id']}}",
        )
        globals_sheet1 = (
            "ID,value\n"
            "message,Goodbye\n"
            "spaced id,This is weird\n"
        )
        globals_sheet2 = (
            "ID,value\n"
            "message,Hello\n"
        )
        minimaldata = (
            "ID,name\n"
            "row1,Name1\n"
            "row2,Name2\n"
        )
        my_template = (
            "row_id,type,from,message_text\n"
            ",send_message,start,{{globals.message}} {{name}}\n"
            ",send_message,,{{message}}\n"
        )
        sheet_dict = {
            "minimaldata": minimaldata,
            "my_template": my_template,
            "my_basic_flow": my_basic_flow,
            "globals_sheet1": globals_sheet1,
            "globals_sheet2": globals_sheet2,
        }
        render_output = (
            ContentIndexParser(
                SheetDataSource(
                    [MockSheetReader(ci_sheet, sheet_dict)]
                ),
                "tests.datarowmodels.minimalmodel",
            )
            .parse_all()
            .render()
        )

        self.assertFlowMessages(
            render_output,
            "my_basic_flow",
            ["Hello", "This is weird"],
        )
        self.assertFlowMessages(
            render_output,
            "my_template - row1",
            ["Hello Name1", "Local message"],
        )

    def test_insert_as_block(self):
        ci_sheet = (
            "type,sheet_name,data_sheet,data_row_id,new_name,data_model,status\n"
            "template_definition,my_template,,,,,\n"
            "create_flow,my_basic_flow,,,my_renamed_basic_flow,,\n"
            "data_sheet,nesteddata,,,,NestedRowModel,\n"
        )
        nesteddata = (
            "ID,value1,custom_field.happy,custom_field.sad\n"
            "row1,Value1,Happy1,Sad1\n"
            "row2,Value2,Happy2,Sad2\n"
        )
        my_template = (
            "row_id,type,from,condition,message_text\n"
            ",send_message,start,,{{value1}}\n"
            "1,wait_for_response,,,\n"
            ",send_message,1,happy,I'm {{custom_field.happy}}\n"
            ",send_message,1,sad,I'm {{custom_field.sad}}\n"
            ",hard_exit,,,\n"
            ",send_message,1,,I'm something\n"
        )
        my_basic_flow = (
            "row_id,type,from,message_text,data_sheet,data_row_id\n"
            ",send_message,start,Some text,,\n"
            "1,insert_as_block,,my_template,nesteddata,row1\n"
            ",send_message,,Next message 1,,\n"
            ",insert_as_block,,my_template,nesteddata,row2\n"
            ",send_message,,Next message 2,,\n"
            ",go_to,,1,,\n"
        )
        sheet_dict = {
            "nesteddata": nesteddata,
            "my_template": my_template,
            "my_basic_flow": my_basic_flow,
        }
        messages_exp = [
            "Some text",
            "Value1",
            "I'm Happy1",
            "Next message 1",
            "Value2",
            "I'm something",
            "Next message 2",
            "Value1",
            "I'm Sad1",  # we're taking the hard exit now, leaving the flow.
        ]
        render_output = (
            ContentIndexParser(
                SheetDataSource([MockSheetReader(ci_sheet, sheet_dict)]),
                "tests.datarowmodels.nestedmodel",
            )
            .parse_all()
            .render()
        )

        self.assertFlowMessages(
            render_output,
            "my_renamed_basic_flow",
            messages_exp,
            Context(inputs=["happy", "else", "sad"]),
        )

    def test_insert_as_block_with_sheet_arguments(self):
        ci_sheet = (
            "type,sheet_name,data_sheet,data_row_id,template_arguments,new_name,data_model,status\n"  # noqa: E501
            "template_definition,my_template,,,lookup;sheet|,,,\n"
            "create_flow,my_template,nesteddata,row3,string_lookup,,,\n"
            "create_flow,my_basic_flow,,,,,,\n"
            "data_sheet,nesteddata,,,,,ListRowModel,\n"
            "data_sheet,string_lookup,,,,,LookupRowModel,\n"
        )
        nesteddata = (
            "ID,messages.1,messages.2\n"
            "row1,hello,nicetosee\n"
            "row2,nicetosee,bye\n"
            "row3,hello,bye\n"
        )
        string_lookup = (
            "ID,happy,sad,neutral\n"
            "hello,Hello :),Hello :(,Hello\n"
            "bye,Bye :),Bye :(,Bye\n"
            "nicetosee,Nice to see you :),Not nice to see you :(,Nice to see you\n"
        )
        my_template = (
            "row_id,type,from,condition,message_text\n"
            "1,split_by_value,,,@field.mood\n"
            ",send_message,1,happy,{% for msg in messages %}{{lookup[msg].happy}}{% endfor %}\n"  # noqa: E501
            ",send_message,1,sad,{% for msg in messages %}{{lookup[msg].sad}}{% endfor %}\n"  # noqa: E501
            ",send_message,1,,{% for msg in messages %}{{lookup[msg].neutral}}{% endfor %}\n"  # noqa: E501
        )
        my_basic_flow = (
            "row_id,type,from,message_text,data_sheet,data_row_id,template_arguments\n"
            ",send_message,start,Some text,,,\n"
            ",insert_as_block,,my_template,nesteddata,row1,string_lookup\n"
            ",send_message,,Intermission,,,\n"
            ",insert_as_block,,my_template,nesteddata,row2,string_lookup\n"
        )
        sheet_dict = {
            "nesteddata": nesteddata,
            "my_template": my_template,
            "my_basic_flow": my_basic_flow,
            "string_lookup": string_lookup,
        }
        render_output = (
            ContentIndexParser(
                SheetDataSource([MockSheetReader(ci_sheet, sheet_dict)]),
                "tests.datarowmodels.listmodel",
            )
            .parse_all()
            .render()
        )

        self.assertFlowMessages(
            render_output,
            "my_basic_flow",
            [
                "Some text",
                "Hello :)Nice to see you :)",
                "Intermission",
                "Nice to see you :)Bye :)",
            ],
            Context(variables={"@field.mood": "happy"}),
        )
        self.assertFlowMessages(
            render_output,
            "my_basic_flow",
            [
                "Some text",
                "Hello :(Not nice to see you :(",
                "Intermission",
                "Not nice to see you :(Bye :(",
            ],
            Context(variables={"@field.mood": "sad"}),
        )
        self.assertFlowMessages(
            render_output,
            "my_basic_flow",
            [
                "Some text",
                "HelloNice to see you",
                "Intermission",
                "Nice to see youBye",
            ],
            Context(variables={"@field.mood": "something else"}),
        )
        self.assertFlowMessages(
            render_output,
            "my_template - row3",
            ["Hello :)Bye :)"],
            Context(variables={"@field.mood": "happy"}),
        )

    def test_insert_as_block_with_arguments(self):
        ci_sheet = (
            "type,sheet_name,data_sheet,data_row_id,template_arguments,new_name,data_model,status\n"  # noqa: E501
            "template_definition,my_template,,,arg1|arg2;;default2,,,\n"
            "create_flow,my_template,,,value1,my_template_default,,\n"
            "create_flow,my_template,,,value1;value2,my_template_explicit,,\n"
        )
        my_template = (
            "row_id,type,from,condition,message_text\n"
            ",send_message,,,{{arg1}} {{arg2}}\n"
        )
        sheet_dict = {
            "my_template": my_template,
        }
        render_output = (
            ContentIndexParser(
                SheetDataSource([MockSheetReader(ci_sheet, sheet_dict)]),
                "tests.datarowmodels.listmodel",
            )
            .parse_all()
            .render()
        )

        self.assertFlowMessages(
            render_output,
            "my_template_default",
            ["value1 default2"],
        )
        self.assertFlowMessages(
            render_output,
            "my_template_explicit",
            ["value1 value2"],
        )

    def test_eval(self):
        ci_sheet = (
            "type,sheet_name,data_sheet,data_row_id,new_name,data_model,template_arguments,status\n"  # noqa: E501
            "template_definition,flow,,,,,metadata;sheet|,\n"
            "data_sheet,content,,,,EvalContentModel,,\n"
            "data_sheet,metadata,,,,EvalMetadataModel,,\n"
            "create_flow,flow,content,,,,metadata,\n"
        )
        metadata = csv_join(
            "ID,include_if",
            "a,text",
        )
        flow = (
            '"row_id","type","from","loop_variable","include_if","message_text"\n'
            ',"send_message",,,,"hello"\n'
            ',"send_message",,,"{@metadata[""a""].include_if|eval == ""yes""@}","{{text}}"\n'  # noqa: E501
        )
        content = csv_join(
            "ID,text",
            "id1,yes",
            "id2,no",
        )
        sheet_dict = {
            "metadata": metadata,
            "content": content,
            "flow": flow,
        }
        render_output = (
            ContentIndexParser(
                SheetDataSource([MockSheetReader(ci_sheet, sheet_dict)]),
                "tests.datarowmodels.evalmodels",
            )
            .parse_all()
            .render()
        )

        self.assertFlowMessages(render_output, "flow - id1", ["hello", "yes"])
        self.assertFlowMessages(render_output, "flow - id2", ["hello"])

    def test_tags(self):
        ci_sheet = (
            "type,sheet_name,new_name,template_arguments,tags.1,tags.2\n"
            "template_definition,flow,,arg,,\n"
            "create_flow,flow,flow-world,World,,\n"
            "create_flow,flow,flow-t1,Tag1Only,tag1,\n"
            "create_flow,flow,flow-b1,Bag1Only,bag1,\n"
            "create_flow,flow,flow-t2,Tag2Only,,tag2\n"
            "create_flow,flow,flow-b2,Bag2Only,,bag2\n"
            "create_flow,flow,flow-t1t2,Tag1Tag2,tag1,tag2\n"
            "create_flow,flow,flow-t1b2,Tag1Bag2,tag1,bag2\n"
            "create_flow,flow,flow-b1t2,Bag1Tag2,bag1,tag2\n"
        )
        flow = (
            '"row_id","type","from","loop_variable","include_if","message_text"\n'
            ',"send_message",,,,"Hello {{arg}}"\n'
        )
        sheet_dict = {
            "flow": flow,
        }
        sheet_reader = MockSheetReader(ci_sheet, sheet_dict)
        render_output = (
            ContentIndexParser(
                SheetDataSource([sheet_reader]), "tests.datarowmodels.evalmodels"
            )
            .parse_all()
            .render()
        )

        self.assertFlowNamesEqual(
            render_output,
            {
                "flow-world",
                "flow-t1",
                "flow-b1",
                "flow-t2",
                "flow-b2",
                "flow-t1t2",
                "flow-t1b2",
                "flow-b1t2",
            },
        )
        self.assertFlowMessages(render_output, "flow-world", ["Hello World"])
        self.assertFlowMessages(render_output, "flow-t1", ["Hello Tag1Only"])
        self.assertFlowMessages(render_output, "flow-b1", ["Hello Bag1Only"])
        self.assertFlowMessages(render_output, "flow-t2", ["Hello Tag2Only"])
        self.assertFlowMessages(render_output, "flow-b2", ["Hello Bag2Only"])
        self.assertFlowMessages(render_output, "flow-t1t2", ["Hello Tag1Tag2"])
        self.assertFlowMessages(render_output, "flow-t1b2", ["Hello Tag1Bag2"])
        self.assertFlowMessages(render_output, "flow-b1t2", ["Hello Bag1Tag2"])

        self.assertFlowNamesEqual(
            ContentIndexParser(
                SheetDataSource([sheet_reader]),
                "tests.datarowmodels.evalmodels",
                TagMatcher(["1", "tag1"]),
            )
            .parse_all()
            .render(),
            {
                "flow-world",
                "flow-t1",
                "flow-t2",
                "flow-b2",
                "flow-t1t2",
                "flow-t1b2",
            },
        )

        self.assertFlowNamesEqual(
            ContentIndexParser(
                SheetDataSource([sheet_reader]),
                "tests.datarowmodels.evalmodels",
                TagMatcher(["1", "tag1", "bag1"]),
            )
            .parse_all()
            .render(),
            {
                "flow-world",
                "flow-t1",
                "flow-b1",
                "flow-t2",
                "flow-b2",
                "flow-t1t2",
                "flow-t1b2",
                "flow-b1t2",
            },
        )

        self.assertFlowNamesEqual(
            ContentIndexParser(
                SheetDataSource([sheet_reader]),
                "tests.datarowmodels.evalmodels",
                TagMatcher(["1", "tag1", "2", "tag2"]),
            )
            .parse_all()
            .render(),
            {
                "flow-world",
                "flow-t1",
                "flow-t2",
                "flow-t1t2",
            },
        )

        self.assertFlowNamesEqual(
            ContentIndexParser(
                SheetDataSource([sheet_reader]),
                "tests.datarowmodels.evalmodels",
                TagMatcher(["5", "tag1", "bag1"]),
            )
            .parse_all()
            .render(),
            {
                "flow-world",
                "flow-t1",
                "flow-b1",
                "flow-t2",
                "flow-b2",
                "flow-t1t2",
                "flow-t1b2",
                "flow-b1t2",
            },
        )

    def assertFlowNamesEqual(self, rapidpro_export: dict, flow_names: Set[str]):
        return self.assertEqual(
            {flow["name"] for flow in rapidpro_export["flows"]},
            flow_names,
        )


class TestOverrideBehaviour(TestCase):

    def test_data_defined_earlier_is_overridden_by_later_definitions(self):
        base = DatasetSheetReader(
            [
                Dataset(
                    ("A", "", ""),
                    ("B", "", ""),
                    headers=("ID", "value1", "value2"),
                    title="data",
                ),
                Dataset(
                    ("data_sheet", "data", "", "SimpleRowModel"),
                    headers=("type", "sheet_name", "new_name", "data_model"),
                    title="content_index",
                ),
            ],
            "base",
        )
        filter_ = DatasetSheetReader(
            [
                Dataset(
                    (
                        "data_sheet",
                        "data",
                        "data",
                        "SimpleRowModel",
                        'filter | expression ; ID != "B"',
                    ),
                    headers=(
                        "type",
                        "sheet_name",
                        "new_name",
                        "data_model",
                        "operation",
                    ),
                    title="content_index",
                )
            ],
            "filter",
        )
        deployment = DatasetSheetReader(
            [
                Dataset(
                    ("B", "", ""),
                    ("C", "", ""),
                    headers=("ID", "value1", "value2"),
                    title="data",
                ),
                Dataset(
                    ("data_sheet", "data", "data", "SimpleRowModel"),
                    headers=("type", "sheet_name", "new_name", "data_model"),
                    title="content_index",
                ),
            ],
            "deployment",
        )
        definition = ContentIndexParser(
            SheetDataSource([base, filter_, deployment]),
            "tests.datarowmodels.simplemodel",
        ).definition
        keys = list(definition.get_data_sheet_rows("data").keys())

        self.assertEqual(keys, ["B", "C"], "Deployment dataset should prevail")


class TestConcatOperation(TestCase):
    def setUp(self):
        simpleA = csv_join(
            "ID,value1,value2",
            "rowA,1A,2A",
        )
        simpleB = csv_join(
            "ID,value1,value2",
            "rowB,1B,2B",
        )
        self.sheet_dict = {
            "simpleA": simpleA,
            "simpleB": simpleB,
        }

    def test_two_fresh_sheets(self):
        self.ci_sheet = csv_join(
            "type,sheet_name,data_sheet,data_row_id,new_name,data_model,operation.type",
            "data_sheet,simpleA;simpleB,,,simpledata,SimpleRowModel,concat",
        )
        self.check_concat()

    def test_two_fresh_sheets_implictly(self):
        self.ci_sheet = csv_join(
            "type,sheet_name,data_sheet,data_row_id,new_name,data_model,operation.type",
            "data_sheet,simpleA;simpleB,,,simpledata,SimpleRowModel,",
        )
        self.check_concat()

    def test_fresh_and_existing_sheets(self):
        self.ci_sheet = csv_join(
            "type,sheet_name,data_sheet,data_row_id,new_name,data_model,operation.type",
            "data_sheet,simpleA,,,renamedA,SimpleRowModel,",
            "data_sheet,renamedA;simpleB,,,simpledata,SimpleRowModel,concat",
        )
        self.check_concat()

    def test_two_existing_sheets(self):
        self.ci_sheet = csv_join(
            "type,sheet_name,data_sheet,data_row_id,new_name,data_model,operation.type",
            "data_sheet,simpleA,,,renamedA,SimpleRowModel,\n"
            "data_sheet,simpleB,,,renamedB,SimpleRowModel,\n"
            "data_sheet,renamedA;renamedB,,,simpledata,SimpleRowModel,concat\n",
        )
        self.check_concat()

    def check_concat(self):
        definition = ContentIndexParser(
            SheetDataSource([MockSheetReader(self.ci_sheet, self.sheet_dict)]),
            "tests.datarowmodels.simplemodel",
        ).definition
        datamodelA = definition.get_data_sheet_row("simpledata", "rowA")
        datamodelB = definition.get_data_sheet_row("simpledata", "rowB")

        self.assertEqual(datamodelA.value1, "1A")
        self.assertEqual(datamodelA.value2, "2A")
        self.assertEqual(datamodelB.value1, "1B")
        self.assertEqual(datamodelB.value2, "2B")


class TestOperation(TestCase):
    def setUp(self):
        self.simple = csv_join(
            "ID,value1,value2",
            "rowA,orange,fruit",
            "rowB,potato,root",
            "rowC,apple,fruit",
            "rowD,Manioc,root",
        )

    def test_filter_fresh(self):
        # The filter operation is referencing a sheet new (not previously parsed) sheet
        self.ci_sheet = (
            "type,sheet_name,data_sheet,data_row_id,new_name,data_model,operation\n"
            "data_sheet,simpleA,,,simpledata,SimpleRowModel,filter|expression;value2=='fruit'\n"  # noqa: E501
        )

        self.create_parser()

        self.assertRowsExistInOrder(["rowA", "rowC"])
        self.assertRowContent()

    def test_filter_existing(self):
        # The filter operation is referencing a previously parsed sheet
        self.ci_sheet = (
            "type,sheet_name,data_sheet,data_row_id,new_name,data_model,operation\n"
            "data_sheet,simpleA,,,,SimpleRowModel,\n"
            "data_sheet,simpleA,,,simpledata,SimpleRowModel,filter|expression;value2=='fruit'\n"  # noqa: E501
        )

        self.create_parser()

        self.assertRowsExistInOrder(["rowA", "rowC"])
        self.assertRowContent()
        self.assertOriginalDataNotModified("simpleA")

    def test_filter_existing_renamed(self):
        self.ci_sheet = (
            "type,sheet_name,data_sheet,data_row_id,new_name,data_model,operation\n"
            "data_sheet,simpleA,,,renamedA,SimpleRowModel,\n"
            "data_sheet,renamedA,,,simpledata,SimpleRowModel,filter|expression;value2=='fruit'\n"  # noqa: E501
        )
        self.create_parser()

        self.assertRowsExistInOrder(["rowA", "rowC"])
        self.assertRowContent()
        self.assertOriginalDataNotModified("renamedA")

    def test_filter_fresh2(self):
        self.ci_sheet = (
            "type,sheet_name,data_sheet,data_row_id,new_name,data_model,operation\n"
            "data_sheet,simpleA,,,simpledata,SimpleRowModel,\"filter|expression;value1 in ['orange','apple']\"\n"  # noqa: E501
        )
        self.create_parser()

        self.assertRowsExistInOrder(["rowA", "rowC"])

    def test_filter_fresh3(self):
        self.ci_sheet = (
            "type,sheet_name,data_sheet,data_row_id,new_name,data_model,operation\n"
            "data_sheet,simpleA,,,simpledata,SimpleRowModel,filter|expression;value1.lower() > 'd'\n"  # noqa: E501
        )

        self.create_parser()

        self.assertRowsExistInOrder(["rowA", "rowB", "rowD"])

    def test_sort(self):
        self.ci_sheet = (
            "type,sheet_name,data_sheet,data_row_id,new_name,data_model,operation\n"
            "data_sheet,simpleA,,,simpledata,SimpleRowModel,sort|expression;value1.lower()\n"  # noqa: E501
        )

        self.create_parser()

        self.assertRowsExistInOrder(["rowC", "rowD", "rowA", "rowB"])
        rows = self.ci_parser.definition.get_data_sheet_rows("simpledata")
        self.assertEqual(rows["rowA"].value1, "orange")
        self.assertEqual(rows["rowA"].value2, "fruit")
        self.assertEqual(rows["rowB"].value1, "potato")
        self.assertEqual(rows["rowC"].value1, "apple")
        self.assertEqual(rows["rowD"].value1, "Manioc")

    def test_sort_existing(self):
        self.ci_sheet = (
            "type,sheet_name,data_sheet,data_row_id,new_name,data_model,operation\n"
            "data_sheet,simpleA,,,,SimpleRowModel,\n"
            "data_sheet,simpleA,,,simpledata,SimpleRowModel,sort|expression;value1.lower()\n"  # noqa: E501
        )

        self.create_parser()

        self.assertRowsExistInOrder(["rowC", "rowD", "rowA", "rowB"])
        self.assertOriginalDataNotModified("simpleA")

    def test_sort_descending(self):
        self.ci_sheet = (
            "type,sheet_name,data_sheet,data_row_id,new_name,data_model,operation\n"
            "data_sheet,simpleA,,,simpledata,SimpleRowModel,sort|expression;value1.lower()|order;descending\n"  # noqa: E501
        )

        self.create_parser()

        self.assertRowsExistInOrder(["rowB", "rowA", "rowD", "rowC"])

    def create_parser(self):
        self.ci_parser = ContentIndexParser(
            SheetDataSource([MockSheetReader(self.ci_sheet, {"simpleA": self.simple})]),
            "tests.datarowmodels.simplemodel",
        )

    def assertRowContent(self):
        rows = self.ci_parser.definition.get_data_sheet_rows("simpledata")

        self.assertEqual(rows["rowA"].value1, "orange")
        self.assertEqual(rows["rowA"].value2, "fruit")
        self.assertEqual(rows["rowC"].value1, "apple")
        self.assertEqual(rows["rowC"].value2, "fruit")

    def assertRowsExistInOrder(self, exp_keys):
        rows = self.ci_parser.definition.get_data_sheet_rows("simpledata")

        self.assertEqual(len(rows), len(exp_keys))
        self.assertEqual(list(rows.keys()), exp_keys)

    def assertOriginalDataNotModified(self, name):
        self.assertEqual(
            list(self.ci_parser.definition.get_data_sheet_rows(name).keys()),
            ["rowA", "rowB", "rowC", "rowD"],
        )


class TestModelInference(TestTemplate):
    def setUp(self):
        self.ci_sheet = (
            "type,sheet_name,data_sheet,data_row_id,status\n"
            "template_definition,my_template,,,\n"
            "create_flow,my_template,mydata,,\n"
            "data_sheet,mydata,,,\n"
        )
        self.my_template = (
            "row_id,type,from,message_text\n"
            ",send_message,start,Lst {{lst.0}} {{lst.1}}\n"
            ",send_message,,{{custom_field.happy}} and {{custom_field.sad}}\n"
        )

    def test_model_inference(self):
        self.mydata = (
            "ID,lst.1:int,lst.2:int,custom_field.happy,custom_field.sad\n"
            "row1,0,4,Happy1,Sad1\n"
            "row2,1,5,Happy2,Sad2\n"
        )

        flows = self.render_flows()

        self.assertFlows(flows)

    def test_model_inference_alt(self):
        self.mydata = (
            "ID,lst:List[int],custom_field.happy,custom_field.sad\n"
            "row1,0;4,Happy1,Sad1\n"
            "row2,1;5,Happy2,Sad2\n"
        )

        flows = self.render_flows()

        self.assertFlows(flows)

    def render_flows(self):
        sheet_dict = {
            "mydata": self.mydata,
            "my_template": self.my_template,
        }

        return (
            ContentIndexParser(
                SheetDataSource([MockSheetReader(self.ci_sheet, sheet_dict)])
            )
            .parse_all()
            .render()
        )

    def assertFlows(self, flows):
        self.assertFlowMessages(
            flows,
            "my_template - row1",
            ["Lst 0 4", "Happy1 and Sad1"],
        )
        self.assertFlowMessages(
            flows,
            "my_template - row2",
            ["Lst 1 5", "Happy2 and Sad2"],
        )


class TestParseCampaigns(TestCase):
    def test_parse_flow_campaign(self):
        ci_sheet = (
            "type,sheet_name,new_name,group\n"
            "create_campaign,my_campaign,renamed_campaign,My Group\n"
            "create_flow,my_basic_flow,,\n"
        )
        my_campaign = (
            "offset,unit,event_type,delivery_hour,message,relative_to,start_mode,flow\n"
            "15,H,F,,,Created On,I,my_basic_flow\n"
        )
        my_basic_flow = csv_join(
            "row_id,type,from,message_text",
            ",send_message,start,Some text",
        )
        sheet_reader = MockSheetReader(
            ci_sheet,
            {
                "my_campaign": my_campaign,
                "my_basic_flow": my_basic_flow,
            },
        )

        render_output = (
            ContentIndexParser(SheetDataSource([sheet_reader])).parse_all().render()
        )

        self.assertEqual(render_output["campaigns"][0]["name"], "renamed_campaign")
        self.assertEqual(render_output["campaigns"][0]["group"]["name"], "My Group")
        event = render_output["campaigns"][0]["events"][0]
        self.assertEqual(event["offset"], 15)
        self.assertEqual(event["unit"], "H")
        self.assertEqual(event["event_type"], "F")
        self.assertEqual(event["delivery_hour"], -1)
        self.assertEqual(event["message"], None)
        self.assertEqual(
            event["relative_to"],
            {"label": "Created On", "key": "created_on"},
        )
        self.assertEqual(event["start_mode"], "I")
        self.assertEqual(event["flow"]["name"], "my_basic_flow")
        self.assertEqual(event["flow"]["uuid"], render_output["flows"][0]["uuid"])
        self.assertIsNone(event.get("base_language"))

    def test_parse_message_campaign(self):
        ci_sheet = csv_join(
            "type,sheet_name,new_name,group",
            "create_campaign,my_campaign,,My Group",
        )
        my_campaign = (
            "offset,unit,event_type,delivery_hour,message,relative_to,start_mode,flow\n"
            "150,D,M,12,Messagetext,Created On,I,\n"
        )

        render_output = (
            ContentIndexParser(
                SheetDataSource(
                    [MockSheetReader(ci_sheet, {"my_campaign": my_campaign})]
                )
            )
            .parse_all()
            .render()
        )

        self.assertEqual(render_output["campaigns"][0]["name"], "my_campaign")
        event = render_output["campaigns"][0]["events"][0]
        self.assertEqual(event["event_type"], "M")
        self.assertEqual(event["delivery_hour"], 12)
        self.assertEqual(event["message"], {"eng": "Messagetext"})
        self.assertEqual(event["base_language"], "eng")

    def test_duplicate_campaign(self):
        ci_sheet = (
            "type,sheet_name,new_name,group\n"
            "create_campaign,my_campaign,,My Group\n"
            "create_campaign,my_campaign2,my_campaign,My Group\n"
        )
        my_campaign = (
            "offset,unit,event_type,delivery_hour,message,relative_to,start_mode,flow\n"
            "150,D,M,12,Messagetext,Created On,I,\n"
        )
        my_campaign2 = (
            "offset,unit,event_type,delivery_hour,message,relative_to,start_mode,flow\n"
            "150,D,M,6,Messagetext,Created On,I,\n"
        )
        sheet_dict = {
            "my_campaign": my_campaign,
            "my_campaign2": my_campaign2,
        }

        render_output = (
            ContentIndexParser(SheetDataSource([MockSheetReader(ci_sheet, sheet_dict)]))
            .parse_all()
            .render()
        )

        self.assertEqual(render_output["campaigns"][0]["name"], "my_campaign")
        self.assertEqual(render_output["campaigns"][0]["events"][0]["delivery_hour"], 6)

    def test_ignore_campaign(self):
        ci_sheet = (
            "type,sheet_name,new_name,group\n"
            "create_campaign,my_campaign,,My Group\n"
            "create_campaign,my_campaign,my_renamed_campaign,My Group\n"
            "ignore_row,my_campaign,,\n"
        )
        my_campaign = (
            "offset,unit,event_type,delivery_hour,message,relative_to,start_mode,flow\n"
            "150,D,M,12,Messagetext,Created On,I,\n"
        )
        sheet_dict = {
            "my_campaign": my_campaign,
        }

        render_output = (
            ContentIndexParser(SheetDataSource([MockSheetReader(ci_sheet, sheet_dict)]))
            .parse_all()
            .render()
        )

        self.assertEqual(len(render_output["campaigns"]), 1)
        self.assertEqual(render_output["campaigns"][0]["name"], "my_renamed_campaign")


class TestParseTriggers(TestCase):
    def test_parse_triggers(self):
        ci_sheet = (
            "type,sheet_name\n"
            "create_triggers,my_triggers\n"
            "create_flow,my_basic_flow\n"
        )
        my_triggers = (
            "type,keywords,flow,groups,exclude_groups,match_type\n"
            "K,the word,my_basic_flow,My Group,,\n"
            "C,,my_basic_flow,My Group;Other Group,,\n"
            "M,,my_basic_flow,,My Group,\n"
            "K,first;second,my_basic_flow,,,F\n"
        )
        my_basic_flow = csv_join(
            "row_id,type,from,message_text",
            ",send_message,start,Some text",
        )

        render_output = (
            ContentIndexParser(
                SheetDataSource(
                    [
                        MockSheetReader(
                            ci_sheet,
                            {
                                "my_triggers": my_triggers,
                                "my_basic_flow": my_basic_flow,
                            },
                        )
                    ]
                )
            )
            .parse_all()
            .render()
        )

        self.assertEqual(render_output["triggers"][0]["trigger_type"], "K")
        self.assertEqual(render_output["triggers"][1]["trigger_type"], "C")
        self.assertEqual(render_output["triggers"][2]["trigger_type"], "M")
        self.assertEqual(render_output["triggers"][0]["keywords"], ["the word"])
        self.assertEqual(render_output["triggers"][0]["keyword"], "the word")
        self.assertEqual(render_output["triggers"][1]["keywords"], [])
        self.assertIsNone(render_output["triggers"][1]["keyword"])
        self.assertEqual(render_output["triggers"][2]["keywords"], [])
        self.assertIsNone(render_output["triggers"][2]["keyword"])
        self.assertEqual(render_output["triggers"][3]["keywords"], ["first", "second"])
        self.assertEqual(render_output["triggers"][3]["keyword"], "first")
        self.assertEqual(render_output["triggers"][0]["match_type"], "F")
        self.assertEqual(render_output["triggers"][3]["match_type"], "F")
        for i in range(3):
            self.assertIsNone(render_output["triggers"][i]["channel"])
            self.assertEqual(
                render_output["triggers"][i]["flow"]["name"],
                "my_basic_flow",
            )
            self.assertEqual(
                render_output["triggers"][i]["flow"]["uuid"],
                render_output["flows"][0]["uuid"],
            )
        mygroup_uuid = render_output["groups"][0]["uuid"]
        groups0 = render_output["triggers"][0]["groups"]
        groups1 = render_output["triggers"][1]["groups"]
        groups2 = render_output["triggers"][2]["exclude_groups"]
        self.assertEqual(groups0[0]["name"], "My Group")
        self.assertEqual(groups0[0]["uuid"], mygroup_uuid)
        self.assertEqual(groups1[0]["name"], "My Group")
        self.assertEqual(groups1[0]["uuid"], mygroup_uuid)
        self.assertEqual(groups1[1]["name"], "Other Group")
        self.assertEqual(groups1[1]["uuid"], render_output["groups"][1]["uuid"])
        self.assertEqual(groups2[0]["name"], "My Group")
        self.assertEqual(groups2[0]["uuid"], mygroup_uuid)

    def test_parse_triggers_without_flow(self):
        ci_sheet = "type,sheet_name\n" "create_triggers,my_triggers\n"
        my_triggers = (
            "type,keywords,flow,groups,exclude_groups,match_type\n"
            "K,the word,my_basic_flow,My Group,,\n"
        )

        with self.assertRaises(RapidProTriggerError):
            ContentIndexParser(
                SheetDataSource(
                    [MockSheetReader(ci_sheet, {"my_triggers": my_triggers})]
                )
            ).parse_all().render()

    def test_ignore_triggers(self):
        ci_sheet = (
            "type,sheet_name\n"
            "create_triggers,my_triggers\n"
            "ignore_row,my_triggers\n"
        )
        my_triggers = (
            "type,keywords,flow,groups,exclude_groups,match_type\n"
            "K,the word,my_basic_flow,My Group,,\n"
        )
        my_basic_flow = csv_join(
            "row_id,type,from,message_text",
            ",send_message,start,Some text",
        )

        render_output = (
            ContentIndexParser(
                SheetDataSource(
                    [
                        MockSheetReader(
                            ci_sheet,
                            {
                                "my_triggers": my_triggers,
                                "my_basic_flow": my_basic_flow,
                            },
                        )
                    ]
                )
            )
            .parse_all()
            .render()
        )

        self.assertEqual(len(render_output["triggers"]), 0)


class TestParseFromFile(TestTemplate):
    def setUp(self):
        self.input_dir = TESTS_ROOT / "input/example1"

    def test_example1_csv(self):
        flows = (
            ContentIndexParser(
                SheetDataSource([CSVSheetReader(self.input_dir / "csv_workbook")]),
                "tests.input.example1.nestedmodel",
            )
            .parse_all()
            .render()
        )

        self.assertFlows(flows)

    def test_example1_csv_composite(self):
        flows = (
            ContentIndexParser(
                SheetDataSource([CSVSheetReader(self.input_dir / "csv_workbook")]),
                "tests.input.example1.nestedmodel",
            )
            .parse_all()
            .render()
        )

        self.assertFlows(flows)

    def test_example1_xlsx(self):
        flows = (
            ContentIndexParser(
                SheetDataSource(
                    [XLSXSheetReader(self.input_dir / "content_index.xlsx")]
                ),
                "tests.input.example1.nestedmodel",
            )
            .parse_all()
            .render()
        )

        self.assertFlows(flows)

    def test_example1_xlsx_composite(self):
        flows = (
            ContentIndexParser(
                SheetDataSource(
                    [XLSXSheetReader(self.input_dir / "content_index.xlsx")]
                ),
                "tests.input.example1.nestedmodel",
            )
            .parse_all()
            .render()
        )

        self.assertFlows(flows)

    def assertFlows(self, flows):
        self.assertFlowMessages(
            flows,
            "my_basic_flow",
            ["Some text"],
        )
        self.assertFlowMessages(
            flows,
            "my_template - row1",
            ["Value1", "Happy1 and Sad1"],
        )
        self.assertFlowMessages(
            flows,
            "my_template - row2",
            ["Value2", "Happy2 and Sad2"],
        )
        self.assertEqual(
            flows["campaigns"][0]["name"],
            "my_campaign",
        )
        self.assertEqual(
            flows["campaigns"][0]["group"]["name"],
            "My Group",
        )
        self.assertEqual(
            flows["campaigns"][0]["events"][0]["flow"]["name"],
            "my_basic_flow",
        )
        self.assertEqual(
            flows["campaigns"][0]["events"][0]["flow"]["uuid"],
            flows["flows"][2]["uuid"],
        )


class TestMultiFile(TestTemplate):
    def test_minimal(self):
        ci_sheet = csv_join(
            "type,sheet_name",
            "template_definition,template",
        )
        self.run_minimal(ci_sheet)

    def test_minimal_singleindex(self):
        self.run_minimal(ci_sheet=None)

    def run_minimal(self, ci_sheet):
        ci_sheet1 = csv_join(
            "type,sheet_name",
            "create_flow,template",
        )
        template = csv_join(
            "row_id,type,from,message_text",
            ",send_message,start,Hello!",
        )
        sheet_dict2 = {
            "template": template,
        }
        sheet_reader1 = MockSheetReader(ci_sheet1, name="mock_1")
        sheet_reader2 = MockSheetReader(ci_sheet, sheet_dict2, name="mock_2")

        self.assertFlowMessages(
            ContentIndexParser(
                SheetDataSource([sheet_reader1, sheet_reader2])
            )
            .parse_all()
            .render(),
            "template",
            ["Hello!"],
        )
        self.assertFlowMessages(
            ContentIndexParser(SheetDataSource([sheet_reader2, sheet_reader1]))
            .parse_all()
            .render(),
            "template",
            ["Hello!"],
        )

    def test_with_model(self):
        ci_sheet1 = (
            "type,sheet_name,data_sheet,data_model,status\n"
            "template_definition,template,,,\n"
            "data_sheet,names,,NameModel,draft\n"
            "create_flow,template,names,,\n"
        )
        ci_sheet2 = (
            "type,sheet_name,data_sheet,data_model,status\n"
            "data_sheet,names,,NameModel,\n"
            "template_definition,template,,,\n"
            "create_flow,template,names,,draft\n"
        )
        template1 = csv_join(
            "row_id,type,from,message_text",
            ",send_message,start,hello {{name}}",
        )
        template2 = csv_join(
            "row_id,type,from,message_text",
            ",send_message,start,hi {{name}}",
        )
        names = csv_join(
            "ID,name",
            "a,georg",
            "b,chiara",
        )
        names = "ID,name\n" "a,georg\n" "b,chiara\n"
        sheet_dict1 = {
            "template": template1,
            "names": names,
        }
        sheet_dict2 = {
            "template": template2,
        }
        sheet_reader1 = MockSheetReader(ci_sheet1, sheet_dict1, name="mock_1")
        sheet_reader2 = MockSheetReader(ci_sheet2, sheet_dict2, name="mock_2")

        flows = (
            ContentIndexParser(
                SheetDataSource([sheet_reader1, sheet_reader2]),
                user_data_model_module_name="tests.datarowmodels.minimalmodel",
            )
            .parse_all()
            .render()
        )

        self.assertFlowMessages(flows, "template - a", ["hi georg"])
        self.assertFlowMessages(flows, "template - b", ["hi chiara"])

        flows = (
            ContentIndexParser(
                SheetDataSource([sheet_reader2, sheet_reader1]),
                user_data_model_module_name="tests.datarowmodels.minimalmodel",
            )
            .parse_all()
            .render()
        )

        self.assertFlowMessages(flows, "template - a", ["hello georg"])
        self.assertFlowMessages(flows, "template - b", ["hello chiara"])



import copy

from rpft.rapidpro.utils import generate_new_uuid
from rpft.rapidpro.models.common import FlowReference, Group, SystemContactField


class CampaignEvent:
    def __init__(
        self,
        offset,
        unit,
        event_type,
        delivery_hour,
        start_mode,
        uuid=None,
        relative_to=None,
        relative_to_key=None,
        relative_to_label=None,
        message=None,
        flow=None,
        flow_name=None,
        flow_uuid=None,
        base_language=None,
    ):
        self.uuid = uuid if uuid else generate_new_uuid()
        self.offset = offset
        self.unit = unit
        self.event_type = event_type
        self.delivery_hour = delivery_hour
        # dict, keys: language IDs, values: message text
        self.message = message
        self.relative_to = relative_to or SystemContactField(
            relative_to_label,
            relative_to_key,
        )
        self.start_mode = start_mode
        self.flow = flow or FlowReference(flow_name, flow_uuid)
        self.base_language = base_language

        if event_type == "M" and (message is None or base_language is None):
            raise ValueError(
                "CampaignEvent must have a message and base_language if the event_type"
                " is M"
            )

        if event_type == "F" and self.flow is None:
            raise ValueError("CampaignEvent must have a flow if the event_type is F")

    def from_dict(data):
        data_copy = copy.deepcopy(data)
        data_copy["relative_to"] = SystemContactField(
            data_copy["relative_to"]["label"],
            data_copy["relative_to"]["key"],
        )

        if "flow" in data_copy:
            data_copy["flow"] = FlowReference(**data_copy["flow"])

        return CampaignEvent(**data_copy)

    def record_global_uuids(self, uuid_dict):
        if self.flow is not None:
            self.flow.record_uuid(uuid_dict)

    def assign_global_uuids(self, uuid_dict):
        if self.flow is not None:
            self.flow.assign_uuid(uuid_dict)

    def render(self):
        data = {
            "uuid": self.uuid,
            "offset": self.offset,
            "unit": self.unit,
            "event_type": self.event_type,
            "delivery_hour": self.delivery_hour,
            "message": self.message,
            "relative_to": self.relative_to.render(),
            "start_mode": self.start_mode,
        }

        if self.event_type == "F" and self.flow:
            data["flow"] = self.flow.render()

        if self.event_type == "M" and self.base_language:
            data["base_language"] = self.base_language

        return data


class Campaign:
    def __init__(
        self, name, group=None, group_name=None, group_uuid=None, events=None, uuid=None
    ):
        self.name = name
        self.group = group or Group(group_name, group_uuid)
        self.uuid = uuid if uuid else generate_new_uuid()
        self.events = events or []

    def add_event(self, event):
        self.events.append(event)

    def from_dict(data):
        assert "group" in data
        assert "name" in data
        assert "events" in data
        return Campaign(
            uuid=data.get("uuid"),
            name=data["name"],
            group=Group.from_dict(data["group"]),
            events=[CampaignEvent.from_dict(event) for event in data["events"]],
        )

    def record_global_uuids(self, uuid_dict):
        for event in self.events:
            event.record_global_uuids(uuid_dict)
        self.group.record_uuid(uuid_dict)

    def assign_global_uuids(self, uuid_dict):
        for event in self.events:
            event.assign_global_uuids(uuid_dict)
        self.group.assign_uuid(uuid_dict)

    def render(self):
        return {
            "group": self.group.render(),
            "name": self.name,
            "uuid": self.uuid,
            "events": [event.render() for event in self.events],
        }



from unittest import TestCase

from rpft.rapidpro.models.actions import EnterFlowAction
from rpft.rapidpro.models.common import generate_field_key
from rpft.rapidpro.models.exceptions import RapidProActionError


class TestActions(TestCase):

    def test_enter_flow_node(self):
        out = EnterFlowAction(flow_name="test_flow", flow_uuid="fake-uuid").render()

        self.assertEqual(out["type"], "enter_flow")
        self.assertEqual(out["flow"]["name"], "test_flow")
        self.assertEqual(out["flow"]["uuid"], "fake-uuid")


class TestFieldKeyGenerator(TestCase):

    def test_generate_key_from_field_name(self):
        self.assertEqual(
            generate_field_key(" Field Name "),
            "field_name",
        )

    def test_generate_keys_up_to_max_length_limit(self):
        self.assertIsNotNone(generate_field_key("x" * 36))

    def test_fail_if_max_key_length_exceeded(self):
        with self.assertRaises(RapidProActionError):
            generate_field_key("z" * 37)

    def test_fail_if_key_does_not_start_with_letter(self):
        with self.assertRaises(RapidProActionError):
            generate_field_key("1zx")



import re
from abc import ABC, abstractmethod

from rpft.parsers.creation.flowrowmodel import (
    Edge,
    FlowRowModel,
)
from rpft.rapidpro.models.actions import (
    Action,
    CallWebhookAction,
    EnterFlowAction,
    TransferAirtimeAction,
)
from rpft.rapidpro.models.common import Exit, generate_field_key, mangle_string
from rpft.rapidpro.models.routers import RandomRouter, SwitchRouter
from rpft.rapidpro.utils import generate_new_uuid

# TODO: EnterFlowNode and WebhookNode are currently children of BaseNode.
# Ideal class tree of nodes:
# BaseNode
#   BasicNode (allows for actions, only has one exit [no router])
#   SwitchRouterNode (has router and multiple exits; should support actions for its
#   subclasses.)
#     EnterFlowNode
#     WebhookNode
#   RandomRouterNode
# TODO: Make BaseNode an abstract class


class BaseNode(ABC):
    def __init__(
        self,
        uuid=None,
        destination_uuid=None,
        default_exit=None,
        actions=None,
        ui_pos=None,
    ):
        self.uuid = uuid or generate_new_uuid()
        self.actions = actions or []
        self.router = None

        # has_basic_exit denotes if the node should have only one exit
        # This generally occurs in cases where there is no router
        # if has_basic_exit is True, we will render a (very basic) exit which
        # points to default_exit_uuid
        self.has_basic_exit = True
        if default_exit:
            self.default_exit = default_exit
        else:
            self.default_exit = Exit(destination_uuid=destination_uuid or None)
        self.exits = [self.default_exit]
        self.ui_pos = ui_pos

    def from_dict(data, ui_data=None):
        if "router" in data:
            if data["router"]["type"] == "random":
                return RandomRouterNode.from_dict(data, ui_data)
            elif data["router"]["type"] == "switch":
                if data["actions"]:
                    if data["actions"][0]["type"] == "enter_flow":
                        return EnterFlowNode.from_dict(data, ui_data)
                    elif data["actions"][0]["type"] == "call_webhook":
                        return CallWebhookNode.from_dict(data, ui_data)
                    elif data["actions"][0]["type"] == "transfer_airtime":
                        return TransferAirtimeNode.from_dict(data, ui_data)
                    else:
                        raise ValueError("Node contains action of invalid type")
                else:
                    return SwitchRouterNode.from_dict(data, ui_data)
            else:
                raise ValueError("Node contains router of invalid type")
        else:
            return BasicNode.from_dict(data, ui_data)

    def add_ui_from_dict(self, ui_dict):
        if self.uuid in ui_dict:
            pos_dict = ui_dict[self.uuid]["position"]
            self.ui_pos = (pos_dict["left"], pos_dict["top"])
        else:
            self.ui_pos = None

    def update_default_exit(self, destination_uuid):
        # TODO: Think of any caveats to storing a node rather than a UUID
        self.default_exit = Exit(destination_uuid=destination_uuid)
        self.exits = [self.default_exit]

    def _add_exit(self, exit):
        self.exits.append(exit)

    def add_action(self, action):
        self.actions.append(action)

    def record_global_uuids(self, uuid_dict):
        for action in self.actions:
            action.record_global_uuids(uuid_dict)

    def assign_global_uuids(self, uuid_dict):
        for action in self.actions:
            action.assign_global_uuids(uuid_dict)

    def add_choice(self):
        raise NotImplementedError

    def validate(self):
        raise NotImplementedError

    def get_last_action(self):
        try:
            return self.actions[-1]
        except IndexError:
            return None

    def get_exits(self):
        return self.exits

    def render(self):
        self.validate()
        # recursively render the elements of the node
        fields = {
            "uuid": self.uuid,
            "exits": [exit.render() for exit in self.get_exits()],
        }
        if self.actions is not None:
            fields["actions"] = [action.render() for action in self.actions]
        if self.router is not None:
            fields.update(
                {
                    "router": self.router.render(),
                }
            )
        return fields

    def render_ui(self):
        if self.ui_pos:
            return {
                "position": {
                    "left": self.ui_pos[0],
                    "top": self.ui_pos[1],
                },
                "type": "execute_actions",
            }
        else:
            return None

    def clear_row_model(self):
        self.row_models = []

    def get_row_models(self):
        return self.row_models

    def prepend_edge_to_row_models(self, edge):
        self.row_models[0].edges.insert(0, edge)

    @abstractmethod
    def short_name(self):
        pass

    @abstractmethod
    def initiate_row_models(self, node_row_id, parent_edge, **kwargs):
        self.row_models = []
        if self.actions:
            for i, action in enumerate(self.actions):
                action_fields = action.get_row_model_fields()
                row_id = f"{node_row_id}.{i}" if i else node_row_id
                row_model = FlowRowModel(
                    row_id=row_id,
                    edges=[parent_edge],
                    node_uuid=self.uuid,
                    ui_position=self.ui_pos or [],
                    **action_fields,
                )
                self.row_models.append(row_model)
                parent_edge = Edge(from_=row_id)
        else:
            self.row_models = [
                FlowRowModel(
                    row_id=node_row_id,
                    edges=[parent_edge],
                    node_uuid=self.uuid,
                    ui_position=self.ui_pos or [],
                    **kwargs,
                )
            ]

    @abstractmethod
    def get_exit_edge_pairs(self):
        pass


class BasicNode(BaseNode):
    # A basic node can accomodate actions and a single (default) exit

    def from_dict(data, ui_data=None):
        exits = [Exit.from_dict(exit_data) for exit_data in data["exits"]]
        if len(exits) != 1:
            raise ValueError("Basic node must have exactly one exit")
        actions = [Action.from_dict(action) for action in data["actions"]]
        return BasicNode(uuid=data["uuid"], default_exit=exits[0], actions=actions)

    def _add_exit(self, exit):
        raise NotImplementedError

    def short_name(self):
        return self.actions[0].short_name()

    def validate(self):
        if not self.has_basic_exit:
            raise ValueError("has_basic_exit must be True for BasicNode")

        if not self.default_exit:
            raise ValueError("default_exit must be set for BasicNode")

    def initiate_row_models(self, current_row_id, parent_edge):
        super().initiate_row_models(current_row_id, parent_edge)

    def get_exit_edge_pairs(self):
        return [(self.default_exit, Edge(from_=self.row_models[-1].row_id))]


class RouterNode(BaseNode, ABC):
    def add_choice(self, *args, **kwargs):
        # Subclasses may choose to validate the input
        self.router.add_choice(*args, **kwargs)

    def update_default_exit(self, destination_uuid):
        self.router.update_default_category(destination_uuid)

    def record_global_uuids(self, uuid_dict):
        super().record_global_uuids(uuid_dict)
        self.router.record_global_uuids(uuid_dict)

    def assign_global_uuids(self, uuid_dict):
        super().assign_global_uuids(uuid_dict)
        self.router.assign_global_uuids(uuid_dict)

    def get_exits(self):
        return self.router.get_exits()

    def get_exit_edge_pairs(self):
        return self.router.get_exit_edge_pairs(self.row_models[-1].row_id)


class SwitchRouterNode(RouterNode):
    def __init__(
        self,
        operand=None,
        result_name=None,
        wait_timeout=None,
        uuid=None,
        router=None,
        ui_pos=None,
    ):
        """
        Either an operand or a router need to be provided
        """
        super().__init__(uuid, ui_pos=ui_pos)
        if router:
            self.router = router
        else:
            if not operand:
                raise ValueError("Either an operand or a router need to be provided")
            self.router = SwitchRouter(operand, result_name, wait_timeout)
        self.has_basic_exit = False

    def from_dict(data, ui_data=None):
        exits = [Exit.from_dict(exit_data) for exit_data in data["exits"]]
        router = SwitchRouter.from_dict(data["router"], exits)
        return SwitchRouterNode(uuid=data["uuid"], router=router)

    def short_name(self):
        short_value = mangle_string(self.router.result_name or self.router.operand)
        if self.router.wait_timeout:
            return f"wait_for.{short_value}"
        else:
            return f"switch.{short_value}"

    def update_no_response_exit(self, destination_uuid):
        self.router.update_no_response_category(destination_uuid)

    def has_wait(self):
        return self.router.has_wait()

    def has_positive_wait(self):
        return self.router.has_positive_wait()

    def validate(self):
        if self.has_basic_exit:
            raise ValueError("Basic exits are not supported in SwitchRouterNode")
        self.router.validate()

    def render_ui(self):
        ui_entry = super().render_ui()
        if not ui_entry:
            return None
        matches_scheme = re.match(
            r'@\(default\(urn_parts\(urns\.([a-z]+)\)\.path,\s+""\)\)',
            self.router.operand,
        )
        # TODO: for results/fields, the title cannot be inferred from the id alone.
        # We should therefore get it using a dict of fields.
        if self.has_wait():
            ui_entry["type"] = "wait_for_response"
            ui_entry["config"] = {"cases": {}}
        elif self.router.operand == "@contact.groups":
            ui_entry["type"] = "split_by_groups"
            ui_entry["config"] = {"cases": {}}
        elif self.router.operand == "@(urn_parts(contact.urn).scheme)":
            ui_entry["type"] = "split_by_scheme"
            ui_entry["config"] = {"cases": {}}
        elif (
            self.router.operand.startswith("@contact.")
            or self.router.operand.startswith("@fields.")
            or matches_scheme
        ):
            if matches_scheme:
                field_id = matches_scheme.group(1)
                field_type = "scheme"
                field_title = field_id.title()
                # TODO: Technically, the field name is custom here.
                # For example, if the id is 'mailto', the name is 'Email Address'
                # As the name is not very important, we ignore this for now.
            else:
                field_id = re.sub("(@contact.)|(@fields.)", "", self.router.operand)
                field_title = field_id
                if self.router.operand.startswith("@contact.") and field_id in [
                    "name",
                    "language",
                    "channel",
                ]:
                    field_type = "property"
                    field_title = field_id.title()
                else:
                    field_type = "field"
            ui_entry["type"] = "split_by_contact_field"
            ui_entry["config"] = {
                "cases": {},
                "operand": {
                    "id": field_id,
                    "type": field_type,  # TODO: can this take other values?
                    "name": field_title,
                },
            }
        elif self.router.operand.startswith("@results."):
            result_id = field_id = re.sub("(@results.)", "", self.router.operand)
            result_title = result_id
            ui_entry["type"] = "split_by_run_result"
            ui_entry["config"] = {
                "cases": {},
                "operand": {
                    "id": result_id,
                    "type": "result",  # TODO: can this take other values?
                    "name": result_title,  # This is a heuristic.
                },
            }
        else:
            ui_entry["type"] = "split_by_expression"
            ui_entry["config"] = {"cases": {}}
        return ui_entry

    def initiate_row_models(self, current_row_id, parent_edge):
        if self.has_wait():
            super().initiate_row_models(
                current_row_id,
                parent_edge,
                type="wait_for_response",
                save_name=self.router.result_name or "",
                no_response=self.router.wait_timeout or "",
            )
        elif (
            self.router.operand == "@contact.groups"
            and self.router.cases
            and self.router.cases[0].type == "has_group"
        ):
            try:
                groups = self.router.cases[0].arguments[1]
            except IndexError:
                groups = ""

            super().initiate_row_models(
                current_row_id,
                parent_edge,
                type="split_by_group",
                mainarg_groups=[groups],
                obj_id=self.router.cases[0].arguments[0] or "",
            )
        else:
            super().initiate_row_models(
                current_row_id,
                parent_edge,
                type="split_by_value",
                mainarg_expression=self.router.operand,
            )


class RandomRouterNode(RouterNode):
    def __init__(self, result_name=None, uuid=None, router=None, ui_pos=None):
        super().__init__(uuid, ui_pos=ui_pos)
        if router:
            self.router = router
        else:
            self.router = RandomRouter(result_name)
        self.has_basic_exit = False

    def from_dict(data, ui_data=None):
        exits = [Exit.from_dict(exit_data) for exit_data in data["exits"]]
        router = RandomRouter.from_dict(data["router"], exits)
        return RandomRouterNode(uuid=data["uuid"], router=router)

    def update_default_exit(self, destination_uuid):
        raise ValueError("RandomRouterNode does not support default exits.")

    def short_name(self):
        if self.router.result_name:
            short_value = mangle_string(self.router.result_name)
            return f"random.{short_value}"
        else:
            return "random"

    def validate(self):
        if self.has_basic_exit:
            raise ValueError("Default exits are not supported in RandomRouterNode")

        self.router.validate()

    def render_ui(self):
        ui_entry = super().render_ui()
        if not ui_entry:
            return None
        ui_entry["type"] = "split_by_random"
        ui_entry["config"] = None
        return ui_entry

    def initiate_row_models(self, current_row_id, parent_edge):
        super().initiate_row_models(
            current_row_id,
            parent_edge,
            type="split_random",
        )


class EnterFlowNode(RouterNode):
    def __init__(
        self,
        flow_name=None,
        flow_uuid=None,
        complete_destination_uuid=None,
        expired_destination_uuid=None,
        uuid=None,
        router=None,
        action=None,
        ui_pos=None,
    ):
        """
        Either an action or a flow_name have to be provided.
        """
        super().__init__(uuid, ui_pos=ui_pos)

        if action:
            if action.type != "enter_flow":
                raise ValueError("Action for EnterFlowNode must be of type enter_flow")
            self.add_action(action)
        else:
            if not flow_name:
                raise ValueError("Either an action or a flow_name have to be provided.")
            self.add_action(EnterFlowAction(flow_name, flow_uuid))

        if router:
            self.router = router
        else:
            self.router = SwitchRouter(
                operand="@child.run.status", result_name=None, wait_timeout=None
            )
            self.router.default_category.update_name("Expired")

            self.add_choice(
                comparison_variable="@child.run.status",
                comparison_type="has_only_text",
                comparison_arguments=["completed"],
                category_name="Complete",
                destination_uuid=complete_destination_uuid,
            )
            self.add_choice(
                comparison_variable="@child.run.status",
                comparison_type="has_only_text",
                comparison_arguments=["expired"],
                category_name="Expired",
                destination_uuid=expired_destination_uuid,
                is_default=True,
            )
            # Suppress the warning about overwriting default category
            self.router.has_explicit_default_category = False

    def from_dict(data, ui_data=None):
        exits = [Exit.from_dict(exit_data) for exit_data in data["exits"]]
        router = SwitchRouter.from_dict(data["router"], exits)
        actions = [Action.from_dict(action) for action in data["actions"]]
        if len(actions) != 1:
            raise ValueError("EnterFlowNode node must have exactly one action")
        return EnterFlowNode(uuid=data["uuid"], router=router, action=actions[0])

    def short_name(self):
        return self.actions[0].short_name()

    def update_default_exit(self, destination_uuid):
        raise ValueError("EnterFlowNode does not support default exits.")

    def update_completed_exit(self, destination_uuid):
        category = self.router._get_category_or_none("Complete")
        category.update_destination_uuid(destination_uuid)

    def update_expired_exit(self, destination_uuid):
        self.router.update_default_category(destination_uuid)

    def validate(self):
        pass

    def render_ui(self):
        ui_entry = super().render_ui()
        if not ui_entry:
            return None
        ui_entry["type"] = "split_by_subflow"
        ui_entry["config"] = {}
        return ui_entry

    def initiate_row_models(self, current_row_id, parent_edge):
        # Note: We want to call the method of the BaseNode class here.
        # If we refactor this to be a child of SwitchRouterNode, careful here!
        super().initiate_row_models(current_row_id, parent_edge)


class CallWebhookNode(RouterNode):
    def __init__(
        self,
        result_name=None,
        url=None,
        method=None,
        body="",
        headers=None,
        success_destination_uuid=None,
        failure_destination_uuid=None,
        uuid=None,
        router=None,
        action=None,
        ui_pos=None,
    ):
        """
        Either an action or a flow_name have to be provided.
        """
        super().__init__(uuid, ui_pos=ui_pos)

        method = method or "POST"
        http_methods = ["CONNECT", "DELETE", "GET", "HEAD", "OPTIONS", "POST", "PUT"]
        if method not in http_methods:
            raise ValueError("Method for WebhookNode must a valid HTTP method")
        headers = headers or {}
        assert type(headers) is dict

        if action:
            if action.type != "call_webhook":
                raise ValueError("Action for WebhookNode must be of type call_webhook")
            self.add_action(action)
        else:
            if not url or not result_name:
                raise ValueError(
                    "Either an action or a url/result_name have to be provided."
                )
            action = CallWebhookAction(
                result_name=result_name,
                url=url,
                method=method,
                body=body,
                headers=headers,
            )
            self.add_action(action)

        if router:
            self.router = router
        else:
            result_field = generate_field_key(result_name)
            self.router = SwitchRouter(
                operand=f"@results.{result_field}.category",
                result_name=None,
                wait_timeout=None,
            )
            self.router.default_category.update_name("Failure")

            self.add_choice(
                comparison_variable=f"@results.{result_field}.category",
                comparison_type="has_only_text",
                comparison_arguments=["Success"],
                category_name="Success",
                destination_uuid=success_destination_uuid,
            )
            self.router.update_default_category(failure_destination_uuid, "Failure")
            # Suppress the warning about overwriting default category
            self.router.has_explicit_default_category = False

    def from_dict(data, ui_data=None):
        exits = [Exit.from_dict(exit_data) for exit_data in data["exits"]]
        router = SwitchRouter.from_dict(data["router"], exits)
        actions = [Action.from_dict(action) for action in data["actions"]]
        if len(actions) != 1:
            raise ValueError("WebhookNode node must have exactly one action")
        return CallWebhookNode(uuid=data["uuid"], router=router, action=actions[0])

    def update_success_exit(self, destination_uuid):
        category = self.router._get_category_or_none("Success")
        category.update_destination_uuid(destination_uuid)

    def update_failure_exit(self, destination_uuid):
        self.update_default_exit(destination_uuid)

    def short_name(self):
        short_value = mangle_string(self.router.result_name or self.actions[0].url)
        return f"webhook.{short_value}"

    def validate(self):
        pass

    def render_ui(self):
        ui_entry = super().render_ui()
        if not ui_entry:
            return None
        ui_entry["type"] = "split_by_webhook"
        ui_entry["config"] = {}
        return ui_entry

    def initiate_row_models(self, current_row_id, parent_edge):
        super().initiate_row_models(current_row_id, parent_edge)


class TransferAirtimeNode(RouterNode):
    def __init__(
        self,
        amounts=None,
        result_name=None,
        success_destination_uuid=None,
        failure_destination_uuid=None,
        uuid=None,
        router=None,
        action=None,
        ui_pos=None,
    ):
        """
        Either an action or a flow_name have to be provided.
        """
        super().__init__(uuid, ui_pos=ui_pos)

        if action:
            if action.type != "transfer_airtime":
                raise ValueError(
                    "Action for TransferAirtimeNode must be of type transfer_airtime"
                )
            self.add_action(action)
        else:
            if not amounts or not result_name:
                raise ValueError(
                    "Either an action or a amounts/result_name have to be provided."
                )
            assert type(amounts) is dict
            action = TransferAirtimeAction(
                amounts=amounts,
                result_name=result_name,
            )
            self.add_action(action)

        if router:
            self.router = router
        else:
            result_field = generate_field_key(result_name)
            self.router = SwitchRouter(
                operand=f"@results.{result_field}",
            )
            self.router.default_category.update_name("Failure")

            self.add_choice(
                comparison_variable=f"@results.{result_field}",
                comparison_type="has_category",
                comparison_arguments=["Success"],
                category_name="Success",
                destination_uuid=success_destination_uuid,
            )
            self.router.update_default_category(failure_destination_uuid, "Failure")
            # Suppress the warning about overwriting default category
            self.router.has_explicit_default_category = False

    def from_dict(data, ui_data=None):
        exits = [Exit.from_dict(exit_data) for exit_data in data["exits"]]
        router = SwitchRouter.from_dict(data["router"], exits)
        actions = [Action.from_dict(action) for action in data["actions"]]
        if len(actions) != 1:
            raise ValueError("TransferAirtimeNode node must have exactly one action")
        return TransferAirtimeNode(uuid=data["uuid"], router=router, action=actions[0])

    def update_success_exit(self, destination_uuid):
        category = self.router._get_category_or_none("Success")
        category.update_destination_uuid(destination_uuid)

    def update_failure_exit(self, destination_uuid):
        self.update_default_exit(destination_uuid)

    def short_name(self):
        short_value = mangle_string(self.actions[0].result_name)
        return f"airtime.{short_value}"

    def validate(self):
        pass

    def render_ui(self):
        ui_entry = super().render_ui()
        if not ui_entry:
            return None
        ui_entry["type"] = "split_by_airtime"
        ui_entry["config"] = {}
        return ui_entry

    def initiate_row_models(self, current_row_id, parent_edge):
        super().initiate_row_models(current_row_id, parent_edge)
